

// ----------------------------------------------------- INCLUDES
#include "./USB/usb.h"
#include "./USB/usb_function_hid.h"
//#include "GenericTypeDefs.h"

#include <math.h>
#include "./common.h"

// ----------------------------------------------------- DEFINES
#define PIN_1  _RA0
#define PIN_2  _RA1
#define PIN_3  _RB0
#define PIN_4  _RB1
#define PIN_5  _RB2
#define PIN_6  _RB3

// ----------------------------------------------------- VARS
extern unsigned char ReceivedDataBuffer[];
extern unsigned char ToSendDataBuffer[];

extern uint8  PinType[MAX_SLAVES][MAX_PINS];

uint16 MasterPinValue[MASTER_PINS_COUNT];
uint16 PeriodHi;



// ==============================================================================================
//   INT2 - HANDLER  -  used for PERIOD, PERIOD_PU and USOUND_SENSOR
// ==============================================================================================

uint8  PeriodUsoundPin;
uint16 Timer2Loops;

void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _INT2Interrupt(void)
//void __attribute__((__interrupt__)) _INT2Interrupt(void)
{
	if (PinType[0][PeriodUsoundPin] == USOUND_SENSOR)
	{
		// ---------------------------------------------------------- USOUND_SENSOR - SRF05
		MasterPinValue[PeriodUsoundPin] = TMR2;
		switch (PeriodUsoundPin)
		{
			case 0: PIN_1 = 0; _TRISA0 = 0; break; // pin to GND and pin in output
			case 1: PIN_2 = 0; _TRISA1 = 0; break; // pin to GND and pin in output
			case 2: PIN_3 = 0; _TRISB0 = 0; break; // pin to GND and pin in output
			case 3: PIN_4 = 0; _TRISB1 = 0; break; // pin to GND and pin in output
			case 4: PIN_5 = 0; _TRISB2 = 0; break; // pin to GND and pin in output
			case 5: PIN_6 = 0; _TRISB3 = 0; break; // pin to GND and pin in output
		}	
		_INT2IE = 0; // disable RP pin interrupt	
	}
	else
	{
		// ---------------------------------------------------------- PERIOD and PERIOD_PU
		PeriodHi = Timer2Loops;
		MasterPinValue[PeriodUsoundPin] = TMR2;
		TMR2 = 0;
		Timer2Loops = 0;	
	}
	//
	_INT2IF = 0;	
}	

// ==============================================================================================
//   TIMER2 INT HANDLER  -  used for PERIOD and PERIOD_PU
// ==============================================================================================

void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _T2Interrupt(void)
//void __attribute__((__interrupt__)) _T2Interrupt(void)
{	
	// ------------------------------------------------
	Timer2Loops++;

	// ------------------------------------------------ reset the interrupt flag
	_T2IF = 0;
}


// ==============================================================================================
//   INTERRUPT - On-Change int handler
// ==============================================================================================

#define CN_BUF_LENGTH 64
uint16 oldPortA;
uint16 oldPortB;
uint16 bufPortA[CN_BUF_LENGTH];
uint16 bufPortB[CN_BUF_LENGTH];
uint8 bufWrIndex = 0;
uint8 bufRdIndex = 0;
uint16 bufMaskA;
uint16 bufMaskB;

void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _CNInterrupt(void)
//void __attribute__((__interrupt__)) _CNInterrupt(void)
{	
	bufPortA[bufWrIndex] = (oldPortA ^ PORTA) & PORTA;
	bufPortB[bufWrIndex] = (oldPortB ^ PORTB) & PORTB;
	if (++bufWrIndex >= CN_BUF_LENGTH) bufWrIndex = 0;
	oldPortA = PORTA;
	oldPortB = PORTB;
	
	_CNIF = 0;
}



// ==============================================================================================
//   TIMER3 INT HANDLER - SERVO and PWM Timed signals - ( 16MHz ) 
// ==============================================================================================

uint8 	BankReadyToSwap;
uint8   EventBank;
uint8	EventIndex;

uint8   EventCount;
uint8   EventPin[MASTER_PINS_COUNT * 8];  // 48 = max 6 pwm
uint32  EventTime[MASTER_PINS_COUNT * 8]; //

uint8 	DeltaPin  [2][MASTER_PINS_COUNT * 8 + 1]; // 49 = max 6 pwm 
uint16	DeltaTime [2][MASTER_PINS_COUNT * 8 + 1]; //


void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _T3Interrupt(void)
//void __attribute__((__interrupt__)) _T3Interrupt(void)
{
	uint8  Pin;
	uint16 TimeToNextEvent;
	uint8  Level;
	do
	{
	 	Pin = 				DeltaPin[EventBank][EventIndex];  // <--- PINS IN EVENTS ARE 1 BASED ---
	 	TimeToNextEvent = 	DeltaTime[EventBank][EventIndex];
	 	
	 	// --------------------------------------------------- test only "reset"
	 	if (Pin == 128)
	 	// --------------------------------------------------- test only DeltaCount ( -1 ? )
	 	//if (EventIndex >= DeltaCount[EventBank] - 1)
	 	// --------------------------------------------------- test "reset" and DeltaCount ( -1 ? )		
	 	//if (Pin == 128 || 
	 	//		EventIndex >= DeltaCount[EventBank] - 1)
	 	{
	 		EventIndex = 0;
	 		
	 		if (BankReadyToSwap)
	 		{
		 		BankReadyToSwap = 0;
		 		EventBank ^= 1;
		 	}	
		 	
		 	break;
		}
		else
		{
			EventIndex++;
			
			Level = Pin >> 4;
			
			switch (Pin & 15)				
			{
				case  1: PIN_1  = Level; break;  // <--- PINS IN EVENTS ARE 1 BASED ---
				case  2: PIN_2  = Level; break;
				case  3: PIN_3  = Level; break;
				case  4: PIN_4  = Level; break;
				case  5: PIN_5  = Level; break;
				case  6: PIN_6  = Level; break;
			}
		}
	}	
	while (TimeToNextEvent < 10);  // <<< if multiple events then intermediate TimeToNextEvent are loosed
	
	// ------------------------------------------------ reset the interrupt flag
	_T3IF = 0;
	// ------------------------------------------------ set the new time to wait
	TMR3 = 65535 - TimeToNextEvent;
}





// ==============================================================================================
//   SERVO and PWM Timed signals  -  Event arrays preparation
// ==============================================================================================
//   
//  1) each ServoValue[] generates two events UP/DW --- EventPin[] (uint8) / EventTime[] (uint32)
//  2) EventTime[] times ( and EventPin[] ) are sorted
//  3) EventPin and EventTime arrays are transformed to delta-times in Pins and Delta arrays
//     ( up to 3 extra events are added to complete the 16 mS time )
//
// ==============================================================================================


// ----------------------------------- wait 12 uS before starting the next servo ( OK )
void GenerateServoEvents()
{
	uint32 time = 0;
	uint8 i;
	for (i = 0; i < MASTER_PINS_COUNT; i++)
	{
		if (PinType[0][i] != SERVO_8 && PinType[0][i] != SERVO_16) continue; // SERVO only
			
		if (MasterPinValue[i] < 1)
		{	
			
			// Removed after RobotArm problems (servo moving when passing from WORK to SLEEP)
			// For PWM signals maybe this is needed to set outputs to OFF (zero)
			// But for SERVO signals it is better stop sending the servo signal
			//  without changing the last output state 
			// ==============================================================================
			//switch (i)
			//{
			//	case  0: PIN_1  = 0; break;
			//	case  1: PIN_2  = 0; break;
			//	case  2: PIN_3  = 0; break;
			//	case  3: PIN_4  = 0; break;
			//	case  4: PIN_5  = 0; break;
			//	case  5: PIN_6  = 0; break;
			//}
			// ------------------------------------------------- ensure OFF 
			//EventPin[EventCount]  = i + 1;		// DOWN
			//EventTime[EventCount] = time;			// DOWN time
			//EventCount++;
			// ==============================================================================
			
			continue;
		}

		EventPin[EventCount]  = i + 1 + 16;			// UP
		EventTime[EventCount] = time;				// UP time
		EventCount++;
		
		// ----------------------------------- 0 mS to 4 mS
		time += MasterPinValue[i];
		
		EventPin[EventCount]  = i + 1;				// DOWN
		EventTime[EventCount] = time;				// DOWN time
		EventCount++;
		
		// ----------------------------------- wait about 12 uS before starting the next servo 
		time += 200;
	}	
}






// ----------------------------------- adaptive start-time

//#define MIN_TIME 10
//
//uint8 Near(uint32 n1, uint32 n2)
//{
//	if (n1 > n2)
//		n1 = n1 - n2;
//	else
//		n1 = n2 - n1;
//
//	return (n1 < MIN_TIME);
//}	

//void GenerateServoEvents()
//{
//	uint32 TimeStart;
//	uint32 TimeEnd;
//	uint8 i;
//	uint8 j;
//	for (i = 0; i < MASTER_PINS_COUNT; i++)
//	{
//		if (ServoValue[i] < MIN_TIME) continue;
//		
//		TimeStart = 0;
//		TimeEnd = ServoValue[i]; 					// 0 mS to 4 mS
//		
//		do
//		{
//			for (j = 0; j < EventCount; j++)
//			{
//				if (Near(TimeStart, EventTime[j]) || Near(TimeEnd, EventTime[j]))
//				{
//					TimeStart += MIN_TIME;
//					TimeEnd += MIN_TIME;
//					break;
//				}
//			}
//		}
//		while (j != EventCount);  // exit only when a free slot is found
//
//		EventPin[EventCount]  = i + 1 + 16;			// UP
//		EventTime[EventCount] = TimeStart;			// UP time
//		EventCount++;
//		
//		EventPin[EventCount]  = i + 1;				// DOWN
//		EventTime[EventCount] = TimeEnd;			// DOWN time
//		EventCount++;
//	}	
//}





void GeneratePwmEvents()
{
	// If a "Pwm" event is < 5uS near to a servo event
	// 	 - the pwm event must be swapped after the servo event
	//   - eventually this 5uS excess-time can be compensated 
	//      with in the next pwm cycle with a 5 uS subtraction
	
	uint32 time = 0;
	uint8 i;
	for (i = 0; i < MASTER_PINS_COUNT; i++)
	{
		time = i*100;
		uint16 value = MasterPinValue[i];
		
		if (PinType[0][i] != PWM_8 && PinType[0][i] != PWM_16) continue; // PWM only
			
		if (value < 1)
		{	
			switch (i)
			{
				case  0: PIN_1  = 0; break;
				case  1: PIN_2  = 0; break;
				case  2: PIN_3  = 0; break;
				case  3: PIN_4  = 0; break;
				case  4: PIN_5  = 0; break;
				case  5: PIN_6  = 0; break;
			}
			
			// ------------------------------------------------- ensure OFF
			EventPin[EventCount]  = i + 1;			// DOWN
			EventTime[EventCount] = time;			// DOWN time
			EventCount++;
			
			continue;
		}
		
		if (value >= 64000)
		{	
			switch (i)
			{
				case  0: PIN_1  = 1; break;
				case  1: PIN_2  = 1; break;
				case  2: PIN_3  = 1; break;
				case  3: PIN_4  = 1; break;
				case  4: PIN_5  = 1; break;
				case  5: PIN_6  = 1; break;
			}
			
			// ------------------------------------------------- ensure ON
			EventPin[EventCount]  = i + 1 + 16;		// UP
			EventTime[EventCount] = time;			// DOWN time
			EventCount++;
			
			continue;
		}
		
		{
			// -----------------------------------
			EventPin[EventCount]  = i + 1 + 16;			// UP
			EventTime[EventCount] = time;				// UP time
			EventCount++;
			
			// ----------------------------------- 0 mS to 4 mS
			time += value;
			
			EventPin[EventCount]  = i + 1;				// DOWN
			EventTime[EventCount] = time;				// DOWN time
			EventCount++;
		}
		
		if (value > 1)
		{
			time = 65536;
			// -----------------------------------
			EventPin[EventCount]  = i + 1 + 16;			// UP
			EventTime[EventCount] = time;				// UP time
			EventCount++;
			
			// ----------------------------------- 4 mS to 8 mS
			time += value;
			
			EventPin[EventCount]  = i + 1;				// DOWN
			EventTime[EventCount] = time;				// DOWN time
			EventCount++;
		}	
		
		if (value > 8)
		{
			time = 65536 * 2;
			// -----------------------------------
			EventPin[EventCount]  = i + 1 + 16;			// UP
			EventTime[EventCount] = time;				// UP time
			EventCount++;
			
			// ----------------------------------- 8 mS to 12 mS
			time += value;
			
			EventPin[EventCount]  = i + 1;				// DOWN
			EventTime[EventCount] = time;				// DOWN time
			EventCount++;
		}	
		
		if (value > 64)
		{
			time = 65536 * 3;
			// -----------------------------------
			EventPin[EventCount]  = i + 1 + 16;			// UP
			EventTime[EventCount] = time;				// UP time
			EventCount++;
		
			// ----------------------------------- 12 mS to 16 mS
			time += value;
			
			EventPin[EventCount]  = i + 1;				// DOWN
			EventTime[EventCount] = time;				// DOWN time
			EventCount++;
		}		
		
	}
}		



// --------------------------------------------------------
//   Gnome Sort - The simplest sorting algorithm
// --------------------------------------------------------
//  It is "stable" ( preserves the original order )
//  Number of swaps = BubbleSort
//  Number of comparations = BubbleSort-comparations / 2
// --------------------------------------------------------
// It works like a garden gnome sorts a line of flower pots
// 1) Look first and second pot
// 2) If they are in the right order step one pot forward
// 3) Otherwise swap pots and step one pot backwards
// 4) If there is no previous pot, step forwards
// 5) When reached the last pot the work is done
// -------------------------------------------------------- 
void GnomeSortEvents()
{
    uint8  temp8;
    uint32 temp32;
    uint16 pos = 1;
    while (pos < EventCount)
    {
	    if (EventTime[pos] >= EventTime[pos-1])
            pos++;
        else
	    {
            temp8 = EventPin[pos-1];
            EventPin[pos-1] = EventPin[pos];
            EventPin[pos] = temp8;
     
            temp32 = EventTime[pos-1];
            EventTime[pos-1] = EventTime[pos];
            EventTime[pos] = temp32;

            if (pos > 1) pos--;
        }
    }
}




// -----------------------------------------------------------------------
//  Generate delta events   ( 16384 units = 1 mS )
// -----------------------------------------------------------------------
void GenerateDeltaEvents()
{
	// ---------------------------------------- after the last event another time with "0"
	EventTime[EventCount] = 65535;
	
	uint8 AltBank = EventBank ^ 1;
	uint8 n = 0;
	uint8 i;
	for (i = 0; i < EventCount; i++)
	{
		DeltaPin[AltBank][n]  = EventPin[i]; 					// Pin (bits 0 to 3) / UP-DOWN (bit 4)
		DeltaTime[AltBank][n] = EventTime[i+1] - EventTime[i];	// delta to next event
		n++;
	}
	
	// ----------------------------------------- Depending on which 4mS quarter contains the last event,
	//                                           add zero, one, two o three dummy blocks to reach 16 mS total.
	//                                           If blocks are "zero" then a zero-length block is required. 
 	uint8 quarter = EventTime[i-1] / 64000;
 	// ----------------------------------------- increasing quarter by 1 reduces total time by 4 mS
 	//quarter +=1;
 	// ----------------------------
	switch (quarter)
	{
		case 0:
			DeltaPin[AltBank][n]  = 0;								// dummy ( no pin changes )
			DeltaTime[AltBank][n] = 64000;							// another 4 mS 
			n++;	
		case 1:	
			DeltaPin[AltBank][n]  = 0;								// dummy ( no pin changes )
			DeltaTime[AltBank][n] = 64000;							// another 4 mS 
			n++;
		case 2:
			DeltaPin[AltBank][n]  = 0 + 128;						// dummy ( no pin changes ) + Reset
			DeltaTime[AltBank][n] = 64000;							// another 4 mS before the restart	
			n++;
			break; // <<< only this break, please !!!
		case 3:	
		default:	
			DeltaPin[AltBank][n]  = 0 + 128;						// dummy ( no pin changes ) + Reset
			DeltaTime[AltBank][n] = 0;								// immediate restart	
			n++;
	}


	// ----------------------------------------- add one dummy block to reach 16 mS total.

	//	DeltaPin[AltBank][n]  = 0;						// dummy ( no pin changes ) + Reset
	//	DeltaTime[AltBank][n] = 65535;					// another 4 mS before the restart	
	//	n++;
		
	//	DeltaPin[AltBank][n]  = 0+128;					// dummy ( no pin changes ) + Reset
	//	DeltaTime[AltBank][n] = 65535;					// another 4 mS before the restart	
	//	n++;

	//DeltaCount[AltBank] = n;
}



void ComputeTimedOutputsTables()
{
	//LED_ON
	
	// ---------------------------------------------------- complete delta-events generation
	EventCount = 0;
	GenerateServoEvents();
	GeneratePwmEvents();

	if (EventCount > 0)
	{
		GnomeSortEvents();
		GenerateDeltaEvents();
		
		// ---------------------------------------------------- bank ready and, eventually, start interrupt
		BankReadyToSwap = 1;
	
		// ---------------------------------------------------- Enable Timer3 interrupt
		_T3IE = 1;
	}
	else
	{
		// ---------------------------------------------------- Disable Timer3 interrupt
		_T3IE = 0;	
	}	
	
	//LED_OFF
}


void StopTimedOutputs()
{
	_T3IE = 0;
	EventIndex = 0;
	BankReadyToSwap = 0;
	EventBank = 0;
	// ----------------------------------- dummy event ( restart every 4 mS )
	DeltaPin[EventBank][0] 	= 0 + 128;
	DeltaTime[EventBank][0] = 65535;
	//DeltaCount[EventBank] = 1;	
	// ----------------------------------- zero to all outputs
	PIN_1 = 0;
	PIN_2 = 0;
	PIN_3 = 0;
	PIN_4 = 0;
	PIN_5 = 0;
	PIN_6 = 0;
}



// ==============================================================================================
//   PIN CONFIGURATIONS
// ==============================================================================================		

void SetPin(uint8 pin, uint8 isInput, uint8 hasPullup, uint8 intOnChange, uint8 isAdc)
{
	switch (pin)
	{
		case  0: _TRISA0 = isInput; _CN2PUE = hasPullup; _CN2IE = intOnChange; break;
		case  1: _TRISA1 = isInput; _CN3PUE = hasPullup; _CN3IE = intOnChange; break;
		case  2: _TRISB0 = isInput; _CN4PUE = hasPullup; _CN4IE = intOnChange; break;
		case  3: _TRISB1 = isInput; _CN5PUE = hasPullup; _CN5IE = intOnChange; break;
		case  4: _TRISB2 = isInput; _CN6PUE = hasPullup; _CN6IE = intOnChange; break;
		case  5: _TRISB3 = isInput; _CN7PUE = hasPullup; _CN7IE = intOnChange; break;
	}
	
	if (intOnChange)
	{
		switch (pin)
		{
			case  0: bufMaskA |= 1<<0; break;
			case  1: bufMaskA |= 1<<1; break;
			case  2: bufMaskB |= 1<<0; break;
			case  3: bufMaskB |= 1<<1; break;
			case  4: bufMaskB |= 1<<2; break;
			case  5: bufMaskB |= 1<<3; break;
		}
	}
	
	if (isAdc)
	{
		switch (pin)
		{
			case  0: _PCFG0 = 0; break;
			case  1: _PCFG1 = 0; break;
			case  2: _PCFG2 = 0; break;
			case  3: _PCFG3 = 0; break;
			case  4: _PCFG4 = 0; break;
			case  5: _PCFG5 = 0; break;
		}
		_ADON = 1; // adc is on
	}	
}


void SetMasterInOutPins()
{
	uint8 pin;
	
	// -------------------------------------------------------- all pins digital ( not AD )
	AD1PCFG = 0xFFFF;
	
	// -------------------------------------------------------- ADC disabled
	_ADON = 0;
	
	// -------------------------------------------------------- disable PERIOD, PERIOD_PU and USOUND
	PeriodUsoundPin = 255;
		
	// -------------------------------------------------------- reset masks
	bufMaskA = 0;
	bufMaskB = 0;
	
	for (pin = 0; pin < MASTER_PINS_COUNT; pin++)
	{
		// ---------------------------------------------------- pin settings ( TRIS and PULL-UP )
		switch (PinType[0][pin])
		{
			case UNUSED:	 		SetPin(pin, 1, 0, 0, 0); break;
			
			case DIG_OUT:	 		SetPin(pin, 0, 0, 0, 0); break;
			case PWM_8:
			case PWM_16: 			SetPin(pin, 0, 0, 0, 0); break;
			case SERVO_8:
			case SERVO_16: 			SetPin(pin, 0, 0, 0, 0); break;
			
			case DIG_IN: 			SetPin(pin, 1, 0, 0, 0); break;
			case DIG_IN_PU:			SetPin(pin, 1, 1, 0, 0); break;
			
			case ADC_8:
			case ADC_16:			SetPin(pin, 1, 0, 0, 1); break;
			case CAP_8:
			case CAP_16:			SetPin(pin, 1, 0, 0, 1); break;
			case RES_8:
			case RES_16:			SetPin(pin, 1, 0, 0, 1); break;
			
			case COUNTER:			SetPin(pin, 1, 0, 1, 0); break;
			case COUNTER_PU:		SetPin(pin, 1, 1, 1, 0); break;
			
			case FAST_COUNTER:		SetPin(pin, 1, 0, 0, 0); break;
			case FAST_COUNTER_PU:	SetPin(pin, 1, 1, 0, 0); break;
			
			case PERIOD:			SetPin(pin, 1, 0, 0, 0); PeriodUsoundPin = pin; break;
			case PERIOD_PU:			SetPin(pin, 1, 1, 0, 0); PeriodUsoundPin = pin; break;
			
			case USOUND_SENSOR: 	SetPin(pin, 1, 1, 0, 0); PeriodUsoundPin = pin; break;  // sensor SRF05 needs a pullup
			
			case I2C_SDA:			SetPin(pin, 1, 0, 0, 0); break;
			case I2C_SCL:			SetPin(pin, 1, 0, 0, 0); break;
			
			case MCA_8_BYTES:		SetPin(pin, 1, 0, 0, 0); break;
			case MCA_16_BYTES:		SetPin(pin, 1, 0, 0, 0); break;
			case MCA_32_BYTES:		SetPin(pin, 1, 0, 0, 0); break;
		}	
	}
	
	// ---------------------------------------------------------------- ADC settings
    _SSRC = 0b111;	// auto conversion starts after sampling
    //AD1CSSL = 0;	// (default 0) no scanning
    //AD1CON2 = 0;	// (default 0) use MUXA - References to AVss and AVdd
    //_ASAM = 0;	// (default 0) sampling begins when SAMP bit is set


	// ---------------------------------------------------------------- FAST_INPUT on RP pin (ST type) 
	for (pin = 0; pin < MASTER_PINS_COUNT; pin++)
	{	
		switch (PinType[0][pin])
		{
			case FAST_COUNTER:
			case FAST_COUNTER_PU:
				// ---------------------------------------------------- connect timer4 to the RP pin
				__builtin_write_OSCCONL(OSCCON & ~(1<<6)); // unlock PPS registers
				switch (pin)
				{
					case  0: RPINR4bits.T4CKR = 5; break;
					case  1: RPINR4bits.T4CKR = 6; break;
					case  2: RPINR4bits.T4CKR = 0; break;
					case  3: RPINR4bits.T4CKR = 1; break;
					case  4: RPINR4bits.T4CKR = 2; break;
					case  5: RPINR4bits.T4CKR = 3; break;
				}
				__builtin_write_OSCCONL(OSCCON | (1<<6)); // lock PPS registers
			break;
		}
	}


	// ---------------------------------------------------------------- PERIOD and USOUND_SENSOR on INT2 
	if (PeriodUsoundPin < MASTER_PINS_COUNT)
	{
		switch (PinType[0][PeriodUsoundPin])
		{
			case USOUND_SENSOR:
			{
				T2CON = 0b1000000000010000;	// Timer2 - ON - 16bit - from internal clock(16 MHz) - prescaled_8 = 0.5 uS
				_INT2IE = 0; // disable pin interrupt
				_T2IE = 0;   // disable Timer2 interrupt
			}
			break;	
	
			case PERIOD:
			case PERIOD_PU:
			{
				T2CON = 0b1000000000000000;	// Timer2 - ON - 16bit - from internal clock(16 MHz) - Max speed = 62.5 nS
				//
				_INT2EP = 0; // 0 = Interrupt on positive edge 
				_INT2IF = 0; // remove int flag
				_INT2IE = 1; // enable pin interrupt
				//
				TMR2 = 0;
				Timer2Loops = 0;
				_T2IF = 0;	 // remove int flag 
				_T2IE = 1;   // enable Timer2 interrupt
			}
			break;
		}
		
		// ------------------------------------------------------------ connect INT2 to RP pin
		__builtin_write_OSCCONL(OSCCON & ~(1<<6)); // unlock PPS registers
		switch (PeriodUsoundPin)
		{
			case 0: RPINR1bits.INT2R = 5; break;
			case 1: RPINR1bits.INT2R = 6; break;
			case 2: RPINR1bits.INT2R = 0; break;
			case 3: RPINR1bits.INT2R = 1; break;
			case 4: RPINR1bits.INT2R = 2; break;
			case 5: RPINR1bits.INT2R = 3; break;
		}
		__builtin_write_OSCCONL(OSCCON | (1<<6));	// lock PPS registers
	}	
	else
	{	
		T2CON = 0b0000000000000000;	// Timer2 - STOP
		_INT2IE = 0; // disable pin interrupt
		_T2IE = 0;   // disable Timer2 interrupt
		
		// ------------------------------------------------------------ detach INT2 from PERIOD pin
		__builtin_write_OSCCONL(OSCCON & ~(1<<6)); // unlock PPS registers
		RPINR1bits.INT2R = 31;
		__builtin_write_OSCCONL(OSCCON | (1<<6));  // lock PPS registers	
	}		
	

	// ---------------------------------------------------------------- IOC always enabled
	_CNIF = 0;
	_CNIE = 1;
	
	// ---------------------------------------------------------------- reset all values to 0 (unused)
	//	for (pin = 0; pin < MASTER_PINS_COUNT; pin++)
	//	{
	//		uint8 pt = PinType[0][pin];
	//		// --------------------------------------- set output values to 0
	//		//if (pt < 128) // 0..127 "output" 
	//		//{
	//		//	MasterPinValue[pin] = 0;
	//		//}
	//		// --------------------------------------- set input values to 0
	//		//if (pt >= 128) // 128..255 "input" 
	//		//{
	//		//	MasterPinValue[pin] = 0;
	//		//}
	//	}				
}


void SetInitialMasterPinTypes()
{
	PinType[0][0] = UNUSED;
	PinType[0][1] = UNUSED;
	PinType[0][2] = UNUSED;
	PinType[0][3] = UNUSED;
	PinType[0][4] = UNUSED;
	PinType[0][5] = UNUSED;
	SetMasterInOutPins();
}



// ===============================================================================================
//   SendValuesToMasterPins
// ===============================================================================================
void SendValuesToMasterPins(uint8* UsbBufIndex)
{
	uint8 ix = *UsbBufIndex;
	uint8 pin;
	for (pin = 0; pin < MASTER_PINS_COUNT; pin++)
	{	
		switch (PinType[0][pin])
		{
			case DIG_OUT:
				MasterPinValue[pin] = ReceivedDataBuffer[ix++];
				break;
			case PWM_8:
			case SERVO_8:
				MasterPinValue[pin] = ReceivedDataBuffer[ix++] * 256;
				break;
			case PWM_16:
			case SERVO_16:
				MasterPinValue[pin] = ReceivedDataBuffer[ix++] * 256;
				MasterPinValue[pin] += ReceivedDataBuffer[ix++];			
				break;
		}
	}
	*UsbBufIndex = ix;
	
	// -------------------------------------------------- Update DigOut pins
	MasterPins_SyncOutputPins();
	// -------------------------------------------------- Update PWM and SERVO events tables
	ComputeTimedOutputsTables();
}


// ===============================================================================================
//   GetValuesFromMasterPins
// ===============================================================================================
void GetValuesFromMasterPins(uint8* UsbBufIndex)
{
	uint8 ix = *UsbBufIndex;
	
	#if (MASTER_LOGGER)
		ToSendDataBuffer[ix++] = RMS_Value >> 8;
		ToSendDataBuffer[ix++] = RMS_Value;

	#elif (MASTER_SNIFFER)
		uint8 nSniffedBytes = 0;
		uint8 SnifferStatus = 0;

		ix += 2;		// Reserved to nSniffedBytes & SnifferStatus

		// ------------------------------------------------------------
		// Get arrived bytes from circular buffer
		// ------------------------------------------------------------
		if (_OERR) {_OERR = 0; SnifferStatus += 128;}
		//RX_Error = 0;
		//_U1RXIE = 0;

		//static uint8 testbyte = 0;
		
		while (RxFifo_ReadIndex != RxFifo_WriteIndex)
		{
	  		RxFifo_ReadIndex++;
	  		if (RxFifo_ReadIndex >= UART1_RX_FIFO_SIZE) RxFifo_ReadIndex = 0;	// wrap if needed
	  		ToSendDataBuffer[ix++] = RxFifo[RxFifo_ReadIndex];			// copy the character
	  		ToSendDataBuffer[ix++] = SniffedFlags[RxFifo_ReadIndex];	// SniffedFlags
	  		nSniffedBytes++;
	  		if (nSniffedBytes >= 24) {SnifferStatus += 64; break;}
	 	} 		

		ToSendDataBuffer[0] = nSniffedBytes;
		ToSendDataBuffer[1] = SnifferStatus;

	#else

		uint8 pin;
		for (pin = 0; pin < MASTER_PINS_COUNT; pin++)
		{			
			switch (PinType[0][pin])
			{
				case DIG_IN:
				case DIG_IN_PU:
					ToSendDataBuffer[ix++] = (uint8)MasterPinValue[pin];
					break;
				case ADC_8:
				case CAP_8:
				case RES_8:
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 6);
					break;
				case ADC_16:
				case CAP_16:
				case RES_16:
					// ------- Tx2Bytes(PinValue[pin] << 2);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 6);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] << 2);
					break;
				case COUNTER:
				case COUNTER_PU:
					// ------ Tx2Bytes(PinValue[pin]);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 8);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin]);
					break;
				case FAST_COUNTER:
				case FAST_COUNTER_PU:
					// ------ Tx2Bytes(TMR1);
					ToSendDataBuffer[ix++] = (uint8)(TMR4 >> 8);
					ToSendDataBuffer[ix++] = (uint8)(TMR4);
					break;
				case PERIOD:
				case PERIOD_PU:
					// ------ Tx2Bytes(PeriodHi);
					ToSendDataBuffer[ix++] = (uint8)(PeriodHi >> 8);
					ToSendDataBuffer[ix++] = (uint8)(PeriodHi);
					// ------ Tx2Bytes(PinValue[pin]);	
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 8);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin]);
					break;
				case USOUND_SENSOR:
					// ------ Tx2Bytes(PinValue[pin] / 2);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 9);
					ToSendDataBuffer[ix++] = (uint8)(MasterPinValue[pin] >> 1);
					break;
				case I2C_SDA:
					break;
				case I2C_SCL:
					break;
				case MCA_8_BYTES:
				case MCA_16_BYTES:
				case MCA_32_BYTES:
					break;
			}	
		}	
	#endif
	
	*UsbBufIndex = ix;
	
	// -------------------------------------------------- Update DigIn/Adc/Cap/Res pins
	MasterPins_SyncInputPins();
}






// ==============================================================================================
//   MasterPins_SyncOutputPins
// ==============================================================================================

void MasterPins_SyncOutputPins()
{	
	// ---------------------------------------------------------- MasterPins_SyncOutputPins START
	//LED_ON

	// ========================================================================
	//  Digital Out
	// ========================================================================
	uint8 i;
	for (i = 0; i < MASTER_PINS_COUNT; i++)
	{
		switch (PinType[0][i])
		{
			case DIG_OUT:
			{
				switch (i)
				{
					case  0: PIN_1  = MasterPinValue[i]; break;
					case  1: PIN_2  = MasterPinValue[i]; break;
					case  2: PIN_3  = MasterPinValue[i]; break;
					case  3: PIN_4  = MasterPinValue[i]; break;
					case  4: PIN_5  = MasterPinValue[i]; break;
					case  5: PIN_6  = MasterPinValue[i]; break;
				}
			}	
			break;
		}	
	}
	
	// ---------------------------------------------------------- LowPriority END
	//LED_OFF		
}




// ==============================================================================================
//   MasterPins_SyncInputPins
// ==============================================================================================

void MasterPins_SyncInputPins()
{	
	// ---------------------------------------------------------- MasterPins_SyncInputPins START
	//LED_ON
	
	// ========================================================================
	//  USOUND_SENSOR - SRF05
	// ========================================================================

	// ------------------------- 65535 * 0.5 uS = 32 mS  ( 32 mS = 5.5 metres )
	if (PeriodUsoundPin < MASTER_PINS_COUNT && PinType[0][PeriodUsoundPin] == USOUND_SENSOR && _T2IF)   
	{	
		
		//MasterPinValue[PeriodUsoundPin] = TMR2;
		
		switch(PeriodUsoundPin) 
		{
			//        Set pin to VCC    	Wait		   Set to GND   Wait	         Set as input
			case 0:	_TRISA0 = 0; _RA0 = 1;  Delay_us(15); _RA0 = 0;     Delay_us(30);	_TRISA0 = 1;  	break;
			case 1:	_TRISA1 = 0; _RA1 = 1;  Delay_us(15); _RA1 = 0;     Delay_us(30);	_TRISA1 = 1;  	break;
			case 2:	_TRISB0 = 0; _RB0 = 1;  Delay_us(15); _RB0 = 0;     Delay_us(30);	_TRISB0 = 1;  	break;
			case 3:	_TRISB1 = 0; _RB1 = 1;  Delay_us(15); _RB1 = 0;     Delay_us(30);	_TRISB1 = 1; 	break;
			case 4:	_TRISB2 = 0; _RB2 = 1;  Delay_us(15); _RB2 = 0;     Delay_us(30);	_TRISB2 = 1;  	break;
			case 5:	_TRISB3 = 0; _RB3 = 1;	Delay_us(15); _RB3 = 0;     Delay_us(30);	_TRISB3 = 1;  	break;
		}

		_INT2EP = 1; // 1 = Interrupt on negative edge
		_INT2IF = 0; // remove int flag
		_INT2IE = 1; // enable pin interrupt
		
		TMR2 = 0;
		_T2IF = 0;
		
					//MasterPinValue[PeriodUsoundPin] = 8000;
	}

	
	// ========================================================================
	//  COUNTERS PIN-CHANGES from circular buffer to PinValues
	// ========================================================================
	while (bufRdIndex != bufWrIndex)
	{
		uint16 n;
		
		n = bufPortA[bufRdIndex];
		n &= bufMaskA;
		if (n != 0)
		{
			if (n & (1<<0)) MasterPinValue[0]++;	
			if (n & (1<<1)) MasterPinValue[1]++;	
		}	
		
		n = bufPortB[bufRdIndex];
		n &= bufMaskB;
		if (n != 0)
		{
			if (n & (1<<0)) MasterPinValue[2]++;	
			if (n & (1<<1)) MasterPinValue[3]++;	
			if (n & (1<<2)) MasterPinValue[4]++;	
			if (n & (1<<3)) MasterPinValue[5]++;	
		}

		if (++bufRdIndex >= CN_BUF_LENGTH) bufRdIndex = 0;
	}
	
	

	
	// ========================================================================
	//  ADC management
	// ========================================================================
	
	if (_ADON)
	{
		uint8 Pin;
		
		for (Pin = 0; Pin < MASTER_PINS_COUNT; Pin++)
		{

			switch (PinType[0][Pin])
			{
				
				case ADC_8:
				case ADC_16:
				{
					// ------------------------------------------------------ set the AdcChannel
					AD1CHS = Pin;
					
					// ------------------------------------------------------ ensure CTMU current disabled
					_CTMUEN = 0;
					
					// ------------------------------------------------------ dummy conversion to remove crosstalk
					//            |  |||||  ||||||
					AD1CON3 =	0b0000100000000000; // sample 8 TAD / ad clock = 1 * TCY ( 500nS + 750nS )
					_IDISSEN = 1; // ground the current source ( works also with CTMU disabled )
					_SAMP = 1;	  // connect C-Hold and discharge
					while (_SAMP) {};
					while (!_DONE) {};
					_IDISSEN = 0; // release the current source grounding
					
					// ------------------------------------------------------ set the final working TAD
					//            |  |||||  ||||||
					//AD1CON3 =	0b0000000100000001; // sample 1 TAD / ad clock = 2 * TCY ( 125nS + 1.5uS )
					//AD1CON3 =	0b0000001000000001; // sample 2 TAD / ad clock = 2 * TCY ( 250nS + 1.5uS )
					//AD1CON3 =	0b0000001100000001; // sample 3 TAD / ad clock = 2 * TCY ( 375nS + 1.5uS )
					//AD1CON3 =	0b0000010000000001; // sample 4 TAD / ad clock = 2 * TCY ( 500nS + 1.5uS )
					//AD1CON3 =	0b0000011000000001; // sample 6 TAD / ad clock = 2 * TCY ( 750nS + 1.5uS )
					//AD1CON3 =	0b0000100000000001; // sample 8 TAD / ad clock = 2 * TCY ( 1uS + 1.5uS )
					//AD1CON3 =	0b0001000000000001; // sample 16 TAD / ad clock = 2 * TCY ( 2uS + 1.5uS )
					AD1CON3 =	0b0001111100000001; // sample 31 TAD / ad clock = 2 * TCY ( 3.9uS + 1.5uS )
					
					// ------------------------------------------------------ accumulate for 16 conversions
					uint16 acc = 0;
					uint8 i;
					for (i = 0; i < 16; i++)
					{
						// -------------------------------------------------- start conversion
						_SAMP = 1;
						// -------------------------------------------------- wait and read the ADC value
						while (_SAMP) {};
						while (!_DONE) {};
						acc += ADC1BUF0;
					}	
					MasterPinValue[Pin] = acc;
				}
				break;
				
				
				case CAP_8:
				case CAP_16:
				{
					//  Project values 
					// ----------------------------------------------------------------------
					//  start value = 700..800
					//  touched = 300..500 ( - 5 lines )
					//  start capacity about 10pF 
					//  touch additional capacity about 10pF 
					// ----------------------------------------------------------------------
					
					//  CTMU current generator trimming
					// ----------------------------------------------------------------------
					//  3 TAD + 55uA +31
					//  4 TAD + 55uA +24
					//  6 TAD + 55uA +0
					// 16 TAD + 55uA -32
					// 31 TAD + 5.5uA +31	
					// ----------------------------------------------------------------------
	
	
					// ------------------------------------------------------ set the AdcChannel
					AD1CHS = Pin;
					
					// ------------------------------------------------------ enable CTMU current
					_CTMUEN =   1;
					_IRNG =     3; // 3=55uA 2=5.5uA 1=0.55uA
					_ITRIM =   31; // -32 to +31 ( 2% increments up to 62% / about 1uA per step at 55uA )
					_EDG2STAT = 0; // Make sure edge2 is 0
					_EDG1STAT = 1; // Set edge1 - Start Charge
			
					// ------------------------------------------------------ dummy conversion to remove crosstalk
					//            |  |||||  ||||||
					AD1CON3 =	0b0000100000000000; // auto sample 8 TAD / ad clock = 1 * TCY ( 500nS + 750nS )
					_IDISSEN = 1; // ground the current source
					_SAMP = 1;	  // connect C-Hold and discharge for 1250nS
					while (_SAMP) {};
					while (!_DONE) {};
					_IDISSEN = 0; // Do not ground the current source
					
					// ------------------------------------------------------ set the final working TAD
					//			  |  |||||  ||||||
					//AD1CON3 =	0b0000000100000001; // sample 1 TAD / ad clock = 2 * TCY ( 125nS + 1.5uS )
					//AD1CON3 =	0b0000001000000001; // sample 2 TAD / ad clock = 2 * TCY ( 250nS + 1.5uS )
					AD1CON3 =	0b0000001100000001; // sample 3 TAD / ad clock = 2 * TCY ( 375nS + 1.5uS )
					//AD1CON3 =	0b0000010000000001; // sample 4 TAD / ad clock = 2 * TCY ( 500nS + 1.5uS )
					//AD1CON3 =	0b0000010100000001; // sample 5 TAD / ad clock = 2 * TCY ( 625nS + 1.5uS )
					//AD1CON3 =	0b0000011000000001; // sample 6 TAD / ad clock = 2 * TCY ( 750nS + 1.5uS )
					//AD1CON3 =	0b0000011100000001; // sample 7 TAD / ad clock = 2 * TCY ( 875nS + 1.5uS )
					//AD1CON3 =	0b0000100000000001; // sample 8 TAD / ad clock = 2 * TCY ( 1uS + 1.5uS )
					//AD1CON3 =	0b0001000000000001; // sample 16 TAD / ad clock = 2 * TCY ( 2uS + 1.5uS )
					//AD1CON3 =	0b0001111100000001; // sample 31 TAD / ad clock = 2 * TCY ( 3.9uS + 1.5uS )
						
					// ------------------------------------------------------ accumulate for 16 conversions	
					uint16 acc = 0;
					uint8 i;
					for (i = 0; i < MASTER_PINS_COUNT; i++)
					{
						// -------------------------------------------------- fully discharge the sensor capacity
						switch(Pin) 
						{
							//       Ground the sensor   	  Wait			Set as input
							case 0:	_TRISA0 = 0;  _RA0 = 0;   Nop();Nop();	_TRISA0 = 1;  	break;	//	AN0
							case 1:	_TRISA1 = 0;  _RA1 = 0;   Nop();Nop();	_TRISA1 = 1;  	break;	// 	AN1
							case 2:	_TRISB0 = 0;  _RB0 = 0;   Nop();Nop();	_TRISB0 = 1;  	break;	//	AN2
							case 3:	_TRISB1 = 0;  _RB1 = 0;   Nop();Nop();	_TRISB1 = 1; 	break;	//	AN3
							case 4:	_TRISB2 = 0;  _RB2 = 0;   Nop();Nop();	_TRISB2 = 1;  	break;	//	AN4
							case 5:	_TRISB3 = 0;  _RB3 = 0;   Nop();Nop();	_TRISB3 = 1;  	break;	//	AN5
						}
						// -------------------------------------------------- start conversion
						_SAMP = 1;
						// -------------------------------------------------- wait and read the ADC value
						while (_SAMP) {};
						while (!_DONE) {};
						acc += ADC1BUF0;
					}	
					MasterPinValue[Pin] = acc;	
					
					// ------------------------------------------------------ disable CTMU current
					_CTMUEN = 0;
				}
				break;
					
				
				case RES_8:
				case RES_16:
				{
					// ------------------------------------------------------ set the AdcChannel
					AD1CHS = Pin;
					
					// ------------------------------------------------------ enable CTMU current
					_CTMUEN =   1;
					_IRNG =     3; // 3=55uA 2=5.5uA 1=0.55uA
					_ITRIM =   15; // -32 to +31 ( 2% increments up to 62% / about 1uA per step at 55uA )
					_EDG2STAT = 0; // Make sure edge2 is 0
					_EDG1STAT = 1; // Set edge1 - Start Charge
					
					// ------------------------------------------------------ dummy conversion to remove crosstalk
					//            |  |||||  ||||||
					  AD1CON3 =	0b0000100000000000; // auto sample 8 TAD / ad clock = 1 * TCY ( 500nS + 750nS )
					_IDISSEN = 1; // ground the current source
					_SAMP = 1;	  // connect C-Hold and discharge for 1250nS
					while (_SAMP) {};
					while (!_DONE) {};
					_IDISSEN = 0; // Do not ground the current source
					
					// ------------------------------------------------------ set the final working TAD
					//            |  |||||  ||||||
					//AD1CON3 =	0b0000000100000001; // sample 1 TAD / ad clock = 2 * TCY ( 125nS + 1.5uS )
					AD1CON3 =	0b0001111100000001; // sample 31 TAD / ad clock = 2 * TCY ( 3.9uS + 1.5uS )
					
					// ------------------------------------------------------ accumulate for 16 conversions
					uint16 acc = 0;
					uint8 i;
					for (i = 0; i < 16; i++)
					{
						// -------------------------------------------------- start conversion
						_SAMP = 1;
						// -------------------------------------------------- wait and read the ADC value
						while (_SAMP) {};
						while (!_DONE) {};
						acc += ADC1BUF0;
					}	
					MasterPinValue[Pin] = acc;
					
					// ------------------------------------------------------ disable CTMU current
					_CTMUEN = 0;
				}
				break;	
				
			}	
		}
	}
	
	
	// ========================================================================
	//  Digital In / Digital In Pullup
	// ========================================================================
	uint8 i;
	for (i = 0; i < MASTER_PINS_COUNT; i++)
	{
		switch (PinType[0][i])
		{
			case DIG_IN:
			case DIG_IN_PU:
			{
				switch (i)
				{
					case  0: MasterPinValue[i] = PIN_1; break;
					case  1: MasterPinValue[i] = PIN_2; break;
					case  2: MasterPinValue[i] = PIN_3; break;
					case  3: MasterPinValue[i] = PIN_4; break;
					case  4: MasterPinValue[i] = PIN_5; break;
					case  5: MasterPinValue[i] = PIN_6; break;
				}
			}
			break;
		}
	}
	
	//Delay_us(8000);
	
	// ---------------------------------------------------------- LowPriority END
	//LED_OFF		
}

