
// ----------------------------------------------------- INCLUDES
#include "./USB/usb.h"
#include "./USB/usb_function_hid.h"
//#include "GenericTypeDefs.h"

#include <math.h>
#include "./common.h"


// ----------------------------------------------------- VARS
volatile uint8  RxFifo[UART1_RX_FIFO_SIZE];
volatile uint16 RxFifo_WriteIndex;
volatile uint16 RxFifo_ReadIndex;

extern unsigned char ReceivedDataBuffer[];
extern unsigned char ToSendDataBuffer[];

extern uint16 RX_TimeOut;
extern uint16 TX_TimeOut;
extern uint8  RX_Error;

extern uint8  NumSlaves;
extern uint8  NumPhysicalSlaves;
extern uint8  SlaveType[MAX_SLAVES];
extern uint8  PinType[MAX_SLAVES][MAX_PINS];

extern uint8  gSpeed;


// ===============================================================================================
//   UTILITY FUNCTIONS
// ===============================================================================================

uint8 gCrc;
void InitCrc()
{
	gCrc = 0;
}	
uint8 AppendArrayToCrc(uint8* bytes, uint16 nBytes)
{
	uint16 i;
    for (i = 0; i < nBytes; i++)
    {
    	gCrc = (gCrc ^ bytes[i]) + 1;
    }
   	return gCrc;
}
uint8 AppendByteToCrc(uint8 new_byte)
{
    gCrc = (gCrc ^ new_byte) + 1;
   	return gCrc;
}	


void EmptyRxFifo()
{
	uint8 Dummy8;
	Dummy8 = U1RXREG;
	Dummy8 = U1RXREG;
	Dummy8 = U1RXREG;
	Dummy8 = U1RXREG;
	//_OERR = 0;
	
	RxFifo_WriteIndex = 0;
	RxFifo_ReadIndex = 0;
	
	if (_OERR) _OERR = 0;
	RX_Error = 0;
	_U1RXIF = 0;         //clear the UART RX interrupt bit
}




// ===============================================================================================
//   TransmitByte
// ===============================================================================================
void TransmitByte(uint8 value)
{
	// --------------------------------- wait a little time before start transmission
	Delay_us(U1BRG >> 2);
	
	// --------------------------------- enable TX
	_UTXEN = 1;
	
	COMM_DIR = 1;

	// --------------------------------- send the byte
	U1TXREG = value;

	// --------------------------------- wait the TX end ( requires UTXSEL<1:0> = 01 ) or timeout
	TMR1 = 0;
	_U1TXIF = 0;
	while (_U1TXIF == 0 && TMR1 < TX_TimeOut);
	
	// --------------------------------- wait also the stop bit completion or timeout
	TMR1 = 0;
	while (_TRMT == 0 && TMR1 < TX_TimeOut);
	
	// --------------------------------- disable TX
	_UTXEN = 0;

	COMM_DIR = 0;
	
	// --------------------------------- remove echo bits
	EmptyRxFifo();
}




// ==============================================================================================
//   RX
// ==============================================================================================
void checkRxErrorUART1(void) 
{
	//uint8 u8_c;
	//check for errors, reset if detected.
	if (U1STAbits.PERR) 
	{
	 	//u8_c = U1RXREG; //clear error
	 	RX_Error = 1;	// reportError("UART1 parity error\n");
	}
	if (U1STAbits.FERR) 
	{
	 	//u8_c = U1RXREG; //clear error
	 	RX_Error = 1;	// reportError("UART1 framing error\n");
	}
	if (U1STAbits.OERR) 
	{
		U1STAbits.OERR = 0; //clear error
		RX_Error = 1;	// reportError("UART1 overrun error\n");
	}
}

//void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _U1RXInterrupt(void)
//{
//#if (MASTER_SNIFFER)
//	//do
//	//{
//		uint8 Flags = 0;
//
//		#if (SNIFFER_LOOPBACK)
//			if (_LATB13) Flags += 8;
//		#else
//			if (COMM_DIR) Flags += 8;
//		#endif
//
//		if (U1STAbits.PERR) Flags += 1;
//		if (U1STAbits.FERR) Flags += 2;
//		if (U1STAbits.OERR) Flags += 4;
//		SniffedFlags[SniffedIndex] = Flags;							// place flags in the buffer	
//		checkRxErrorUART1(); 										
//		SniffedBytes[SniffedIndex] = (uint8)U1RXREG;				// read char and place it in the buffer
//		SniffedIndex++;   											// increment head pointer
//		if (SniffedIndex >= SNIFFED_SIZE) SniffedIndex = 0;			// wrap if needed
//	//}
//	//while (_URXDA);
//#else
//	//do
//	//{
//		checkRxErrorUART1(); 										
//		RxFifo_WriteIndex++;   												// increment head pointer
//		if (RxFifo_WriteIndex >= UART1_RX_FIFO_SIZE) RxFifo_WriteIndex = 0;	// wrap if needed
//		if (RxFifo_WriteIndex == RxFifo_ReadIndex) RX_Error = 1;			// FIFO overrun!, report error
//		RxFifo[RxFifo_WriteIndex] = (uint8)U1RXREG;							// read char and place it in the buffer
//	//}
//	//while (_URXDA);
//#endif
//
//	_U1RXIF = 0;         //clear the UART RX interrupt bit
//}

void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _U1RXInterrupt(void)
{
	//do
	//{

	#if (MASTER_SNIFFER)
		uint8 Flags = 0;
		#if (SNIFFER_LOOPBACK)
			if (_LATB13) Flags += 8;
		#else
			if (COMM_DIR) Flags += 8;
		#endif

		if (U1STAbits.PERR) Flags += 1;
		if (U1STAbits.FERR) Flags += 2;
		if (U1STAbits.OERR) Flags += 4;
	#endif

	checkRxErrorUART1(); 										
	RxFifo_WriteIndex++;   												// increment head pointer
	if (RxFifo_WriteIndex >= UART1_RX_FIFO_SIZE) RxFifo_WriteIndex = 0;	// wrap if needed
	if (RxFifo_WriteIndex == RxFifo_ReadIndex) RX_Error = 1;			// FIFO overrun!, report error
	RxFifo[RxFifo_WriteIndex] = (uint8)U1RXREG;							// read char and place it in the buffer

	#if (MASTER_SNIFFER)
		SniffedFlags[RxFifo_WriteIndex] = Flags;							// place flags in the buffer	}
	#endif

	//}
	//while (_URXDA);

	_U1RXIF = 0;         //clear the UART RX interrupt bit
}


// ------------------------------------------------------------
// Wait for a byte to be available from UART1 RX and return it
// ------------------------------------------------------------
uint8 UART_GetChar(void)
{
	TMR1 = 0;
	while (RxFifo_WriteIndex == RxFifo_ReadIndex)
	{
		ClrWdt();
		if (TMR1 > RX_TimeOut)
		{
			RX_Error = 1;
			return 0;
		}	
	}		 
  	RxFifo_ReadIndex++;
  	if (RxFifo_ReadIndex >= UART1_RX_FIFO_SIZE) RxFifo_ReadIndex = 0; 	// wrap if needed
  	return RxFifo[RxFifo_ReadIndex]; 									// return the character
}


uint8 ReceiveByte()
{
	// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	#if UART_RECEIVE_INTERRUPT == 0
	
		TMR1 = 0;
		
		// --------------------------------- restore the normal fifo beaviour
		if (_OERR) _OERR = 0;
		//if (_FERR) _FERR = 0;
		RX_Error = 0;
	
		// --------------------------------- wait the RX data available or timeout
		while (_URXDA == 0)
		{
			if (TMR1 > RX_TimeOut)
			{
				RX_Error = 1;
				return 0;
			}	
		}
		
		// --------------------------------- if FRAME-ERROR then wait the auto-baud byte ( 0x55 )
		if (_FERR)
		{
			RX_Error = 1;
			return 0;
		}	

		// --------------------------------- receive value
		uint8 value = U1RXREG;
	#endif
	
	// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	#if UART_RECEIVE_INTERRUPT != 0
		if (_OERR) _OERR = 0;
		//if (_FERR) _FERR = 0;
		RX_Error = 0;
	
		uint8 value = UART_GetChar();
		if (RX_Error) 
		{
			return 0;
		}	
	#endif
	
	// --------------------------------- return the byte
	return value;
}


//uint8 ReceiveByte()
//{
//	TMR1 = 0;
//	
//	// --------------------------------- restore the normal fifo beaviour
//	if (_OERR) _OERR = 0;
//	//if (_FERR) _FERR = 0;
//	RX_Error = 0;
//	
//	// --------------------------------- 
//	//EmptyRxFifo();
//
//	// --------------------------------- wait the RX data available - first method - ( requires URXISEL<1:0> = 00 )
//	// TODO set and test timeout timer 
//	//_U1RXIF = 0;
//	//while (_U1RXIF == 0);
//
//	// --------------------------------- wait the RX data available or timeout
//	while (_URXDA == 0)
//	{
//		if (TMR1 > RX_TimeOut)
//		{
//			RX_Error = 1;
//			return 0;
//		}	
//	}	 	 
//
//
//	// --------------------------------- if FRAME-ERROR then wait the auto-baud byte ( 0x55 )
//	if (_FERR)
//	{
//		RX_Error = 1;
//		return 0;
//	}
//
//
//	// --------------------------------- read amd return the byte 
//	return U1RXREG;
//}


//uint32 Receive3Bytes()
//{
//	uint32 RX_Value;
//
//	// --------------------------------- empty the RX fifo
//	EmptyRxFifo();
//
//	// --------------------------------- receive 3 bytes
//	RX_Value = ReceiveByte();
//	//if (RxError) return;
//	RX_Value <<= 8;
//
//	RX_Value += ReceiveByte();
//	//if (RxError) return;
//	RX_Value <<= 8;
//
//	RX_Value += ReceiveByte();
//	
//	return RX_Value;
//}




// ===============================================================================================
//   SET COM SPEED
// ===============================================================================================                          
void SetComSpeed(uint8 Speed)
{
	uint16 KbitPerSecond;
	
	gSpeed = Speed;

	// ------------------------- set baud generator to HiSpeed
	_BRGH = 1;

	switch (Speed)
	{	
		case 1:	// 1 Kbaud
		{	
			KbitPerSecond = 1;
			break;
		}
		case 2:	// 2 Kbaud
		{	
			KbitPerSecond = 2;
			break;
		}
		case 3:	// 5 Kbaud
		{	
			KbitPerSecond = 5;
			break;
		}	
		case 4:	// 10 Kbaud
		{	
			KbitPerSecond = 10;
			break;
		}	
		case 5:	// 20 Kbaud
		{	
			KbitPerSecond = 20;
			break;
		}	
		default:
		case 6:	// 50 Kbaud
		{	
			KbitPerSecond = 50;
			break;
		}	
		case 7:	// 100 Kbaud
		{	
			KbitPerSecond = 100;
			break;
		}	
		case 8:	// 200 Kbaud
		{	
			KbitPerSecond = 200;
			break;
		}
		case 9:	// 500 Kbaud
		{	
			KbitPerSecond = 500;
			break;
		}	
		case 10: // 1 Mbaud
		{	
			KbitPerSecond = 1000;
			break;
		}	
		case 11: // 2 Mbaud
		{	
			KbitPerSecond = 2000;
			break;
		}	
		case 12: // 4 Mbaud
		{	
			KbitPerSecond = 4000;
			break;
		}		
	}
	// ----------------------------------------------------- calc derivate values
	U1BRG = ( FCY / 4 / 1000) / KbitPerSecond - 1;

	// default com speed "6" ( 20 uS - 50KHz ) ----> U1BRG = 79
	
	//#define TIMEOUT 62500 / 15  //  T1_freq / 15 mS    ( Timer1 freq = 16MHz / 256 = 62500 Hz )
	//RX_TimeOut = (62500 / 100) / KbitPerSecond ;
	
	// Timer1 freq = 16MHz / 8 = 2 MHz
	//RX_TimeOut = 30000 / KbitPerSecond ;  		// about 15 mS at speed = 1 KbitPerSecond
	//if (RX_TimeOut < 40000) RX_TimeOut = 40000;   // minimum 20 mS
	
	//RX_TimeOut = (5000 / 20) * 18;  // timeout = 18 mS
		

	//RX_TimeOut = (5000 / 20) * 16;  // timeout = 16 mS
	
	//RX_TimeOut = 625;  // timeout = 20 mS ERRATED
	
	// =========================================================================
	// timer1 speed = 16MHz / 8 = 0.5uS 
	// 1 byte at speed 1 = 10 cells x 1mS = 10mS
	// =========================================================================
	
	// PPPPP  OK  ( speed >= 3 )
	//RX_TimeOut = 5000;  // timeout = 2.5 mS
	//TX_TimeOut = 5000;  // timeout = 2.5 mS
	
	// PPPPP  NO	
	RX_TimeOut = 25000;  // <<< 25000 x 0.5uS = 12.5 mS
	TX_TimeOut = 25000;  // <<< 25000 x 0.5uS = 12.5 mS	
}


// ===============================================================================================
//   RecognizeSlaves
// ===============================================================================================
void RecognizeSlaves()
{
	uint8 SlaveId;
		
	NumSlaves = 0;
	NumPhysicalSlaves = 0;
	
	uint8 CommandOK = TRUE;
		
	#if (MASTER_LOGGER)
	
		SlaveId = 0;
		NumSlaves = 1;	
		ToSendDataBuffer[SlaveId + 2] = SLAVE_TYPE_LOGGER;
		
	#elif (MASTER_SNIFFER)
	
		SlaveId = 0;
		NumSlaves = 1;	
		ToSendDataBuffer[SlaveId + 2] = SLAVE_TYPE_SNIFFER;
	
	#else
	
		// ------------------------------------------------------ VIRTUAL SLAVE
		StopTimedOutputs();
		SlaveId = 0;
		NumSlaves = 1;	
		ToSendDataBuffer[SlaveId + 2] = SLAVE_TYPE_MASTER_PINS;
	
		// ------------------------------------------------------ PHYSICAL SLAVE RECOGNIZE
		Delay_ms(50);
				
		TransmitByte(CMD_RECOG_START);
		TransmitByte(ReceivedDataBuffer[1]);
		
		InitCrc();
		uint8 crc = AppendArrayToCrc((uint8*)&ReceivedDataBuffer[0], 2);
		TransmitByte(crc);
		
		for (SlaveId = NumSlaves; SlaveId < MAX_SLAVES; SlaveId++)
		{
			// --------------------------------------------- call slave "N" 
			TransmitByte(CMD_RECOG);
			TransmitByte(1);
			TransmitByte(SlaveId);
			
			InitCrc();
			AppendByteToCrc(CMD_RECOG);
			AppendByteToCrc(1);
			uint8 crc = AppendByteToCrc(SlaveId);
			TransmitByte(crc);
			
			// -------------------------------------------- receive slave type
			uint8 RX_Value = ReceiveByte();
			if (RX_Error) break; // if no more slaves then exit for
			crc = ReceiveByte();
			
			InitCrc();
			uint8 mycrc = AppendByteToCrc(RX_Value);
			if (mycrc != crc || RX_Error)
			{
				CommandOK = FALSE;
				SlaveId = 0;
				break;
			}	
			
			NumPhysicalSlaves++;
			
			SlaveType[SlaveId] = RX_Value;
	
			switch (SlaveType[SlaveId])
			{
				case SLAVE_TYPE_CAPSENSOR:
					PinType[SlaveId][0] = UNUSED; 
					break;
				case SLAVE_TYPE_SERVO:
					PinType[SlaveId][0] = UNUSED;
					PinType[SlaveId][1] = UNUSED; 
					PinType[SlaveId][2] = UNUSED; 
					PinType[SlaveId][3] = UNUSED; 
					PinType[SlaveId][4] = UNUSED;
					PinType[SlaveId][5] = UNUSED; 
					PinType[SlaveId][6] = UNUSED;
					PinType[SlaveId][7] = UNUSED; 
					PinType[SlaveId][8] = UNUSED; 
					PinType[SlaveId][9] = UNUSED;
					break;
				case SLAVE_TYPE_GENERIC:
					PinType[SlaveId][0] = UNUSED;
					PinType[SlaveId][1] = UNUSED; 
					PinType[SlaveId][2] = UNUSED; 
					PinType[SlaveId][3] = UNUSED; 
					PinType[SlaveId][4] = UNUSED;
					PinType[SlaveId][5] = UNUSED; 
					PinType[SlaveId][6] = UNUSED;
					PinType[SlaveId][7] = UNUSED; 
					PinType[SlaveId][8] = UNUSED; 
					PinType[SlaveId][9] = UNUSED;
					PinType[SlaveId][10] = UNUSED;
					PinType[SlaveId][11] = UNUSED;
					break;
				case SLAVE_TYPE_INOUT:
					PinType[SlaveId][0] = UNUSED;
					PinType[SlaveId][1] = UNUSED; 
					PinType[SlaveId][2] = UNUSED; 
					PinType[SlaveId][3] = UNUSED; 
					PinType[SlaveId][4] = UNUSED;
					PinType[SlaveId][5] = UNUSED; 
					PinType[SlaveId][6] = UNUSED;
					PinType[SlaveId][7] = UNUSED; 
					PinType[SlaveId][8] = UNUSED; 
					PinType[SlaveId][9] = UNUSED;
					PinType[SlaveId][10] = UNUSED;
					PinType[SlaveId][11] = UNUSED;
					break;
			}
			
			ToSendDataBuffer[SlaveId + 2] = RX_Value;
		}
	
		// -------------------------------------------------------------- 
		NumSlaves = SlaveId;
	#endif

	ToSendDataBuffer[1] = NumSlaves;

	// --------------------------------- remove echo bits
	//EmptyRxFifo();

	// --------------------------------------------------------------- END-RECOGNIZE and RESET by SendSpeed 
	ReceivedDataBuffer[0] = CMD_SPEED;
	ReceivedDataBuffer[1] = gSpeed;
	TransmitComSpeedToPhysicalSlaves();	
	ReceivedDataBuffer[0] = CMD_NOCMD;
	
	// ---------------------------------------------------------------
	if (CommandOK)
		ToSendDataBuffer[0] = 0; 
	else
		ToSendDataBuffer[0] = 1;
}



// ===============================================================================================
//   SendValuesToPhysicalSlave
// ===============================================================================================
void SendValuesToPhysicalSlave(uint8 SlaveId, uint8 nbytes, uint8 offset)
{
	// ----------------------------------------- transmit CommandId / SlaveId / nbytes
	TransmitByte(CMD_SEND_VALUES);
	TransmitByte(SlaveId);
	TransmitByte(nbytes);
	// ----------------------------------------- 
	uint16 i;
	for (i = 0; i < nbytes; i++)
	{
		TransmitByte(ReceivedDataBuffer[offset + i]);
	}
	// ----------------------------------------- 	
	InitCrc();
	AppendByteToCrc(CMD_SEND_VALUES);
	AppendByteToCrc(SlaveId);
	AppendByteToCrc(nbytes);
	uint8 crc = AppendArrayToCrc((uint8*)&ReceivedDataBuffer[offset], nbytes);
	TransmitByte(crc);
	// ----------------------------------------- receive the slaveID
	uint8 CommandOK = TRUE;
	uint8 RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	if (RX_Value != SlaveId) CommandOK = FALSE;
	
	// ----------------------------------------- receive the CRC
	RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	// ----------------------------------------- 
	InitCrc();
	crc = AppendByteToCrc(SlaveId);
	if (RX_Value != crc) CommandOK = FALSE;	
	// ----------------------------------------- 
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR		
}	

// ===============================================================================================
//  GetValuesFromPhysicalSlave
// ===============================================================================================
void GetValuesFromPhysicalSlave(uint8 SlaveId, uint8 nbytes, uint8 offset)
{
	// ----------------------------------------- transmit CommandId / SlaveId / nbytes
	TransmitByte(CMD_GET_VALUES);
	TransmitByte(SlaveId);
	TransmitByte(nbytes);
	// ----------------------------------------- CRC from 3 bytes ( CommandId / SlaveId / nbytes )
	InitCrc();
	AppendByteToCrc(CMD_GET_VALUES);
	AppendByteToCrc(SlaveId);
	uint8 crc = AppendByteToCrc(nbytes);
	TransmitByte(crc);
	// ----------------------------------------- receive bytes
	uint8 CommandOK = TRUE;
	uint16 i;
	for (i = 0; i < nbytes; i++)
	{
		ToSendDataBuffer[i + offset] = ReceiveByte();
		if (RX_Error) CommandOK = FALSE;
	}	
	// ----------------------------------------- receive the slaveID
	uint8 RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	if (RX_Value != SlaveId) CommandOK = FALSE;
	// ----------------------------------------- receive the CRC
	RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	// ----------------------------------------- calc and test the crc
	InitCrc();
	AppendArrayToCrc((uint8*)&ToSendDataBuffer[offset], nbytes);
	crc = AppendByteToCrc(SlaveId);
	if (RX_Value != crc) CommandOK = FALSE;	
	// ----------------------------------------- set response
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR		
}


// ===============================================================================================
//   HOST COMMAND - SendBytesToPhysicalSlave
// ===============================================================================================
void SendBytesToPhysicalSlave()
{
	uint8 SlaveId = ReceivedDataBuffer[1];
	uint8 nbytes = ReceivedDataBuffer[2];
	
	uint8 CommandOK = TRUE;
	// ----------------------------------------- transmit CommandId / SlaveId / nbytes
	TransmitByte(CMD_SEND_BYTES);
	TransmitByte(SlaveId);
	TransmitByte(nbytes);
	// ----------------------------------------- 
	uint16 i;
	for (i = 0; i < nbytes; i++)
	{
		TransmitByte(ReceivedDataBuffer[3 + i]);
	}	
	// ----------------------------------------- 
	InitCrc();
	uint8 crc = AppendArrayToCrc((uint8*)&ReceivedDataBuffer[0], 3 + nbytes);
	TransmitByte(crc);
	// ----------------------------------------- receive the slaveID
	uint8 RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	if (RX_Value != SlaveId) CommandOK = FALSE;
	// ----------------------------------------- receive the CRC
	RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	// ----------------------------------------- 
	InitCrc();
	crc = AppendByteToCrc(SlaveId);
	if (RX_Value != crc) CommandOK = FALSE;	
	// ----------------------------------------- 
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR		
}	


// ===============================================================================================
//   HOST COMMAND - GetBytesFromPhysicalSlave
// ===============================================================================================

void GetBytesFromPhysicalSlave()
{
	uint8 SlaveId = ReceivedDataBuffer[1];
	uint8 nbytes = ReceivedDataBuffer[2];
	
	uint8 CommandOK = TRUE;
	// ----------------------------------------- transmit CommandId / SlaveId / nbytes
	TransmitByte(CMD_GET_BYTES);
	TransmitByte(SlaveId);
	TransmitByte(nbytes);
	// ----------------------------------------- CRC from 3 bytes ( CommandId / SlaveId / nbytes )
	InitCrc();
	uint8 crc = AppendArrayToCrc((uint8*)&ReceivedDataBuffer[0], 3);
	TransmitByte(crc);
	// ----------------------------------------- receive bytes
	uint16 i;
	for (i = 0; i < nbytes; i++)
	{
		ToSendDataBuffer[i + 1] = ReceiveByte();
		if (RX_Error) CommandOK = FALSE;
	}	
	// ----------------------------------------- receive the slaveID
	uint8 RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	if (RX_Value != SlaveId) CommandOK = FALSE;
	// ----------------------------------------- receive the CRC
	RX_Value = ReceiveByte();
	if (RX_Error) CommandOK = FALSE;
	// ----------------------------------------- calc and test the crc
	InitCrc();
	ToSendDataBuffer[nbytes + 1] = SlaveId;
	crc = AppendArrayToCrc((uint8*)&ToSendDataBuffer[1], nbytes + 1);
	if (RX_Value != crc) CommandOK = FALSE;	
	// ----------------------------------------- set response
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR		
}



// ===============================================================================================
//   HOST COMMAND - TransmitComSpeedToPhysicalSlaves
// ===============================================================================================

void TransmitComSpeedToPhysicalSlaves()
{
	uint8 Speed 	= ReceivedDataBuffer[1];
	InitCrc();
	uint8 Crc 		= AppendArrayToCrc((uint8*)&ReceivedDataBuffer[0], 2);
					
	// --------------------------------- wait a long time
	Delay_ms(50);

	// --------------------------------- set the minimal speed
	SetComSpeed(1);

	// --------------------------------- enable TX
	_UTXEN = 1;

	COMM_DIR = 1;

	// --------------------------------- send break to slaves
	_UTXBRK = 1;
	U1TXREG = 0;

	// --------------------------------- wait the completion of break character
	while (_UTXBRK == 1);
	Delay_ms(10);

	// --------------------------------- send AUTO-BAUD character
	SetComSpeed(Speed);
	U1TXREG = 0x55;

	// --------------------------------- wait the TX end ( requires UTXSEL<1:0> = 01 )
	_U1TXIF = 0;
	while (_U1TXIF == 0);
	// --------------------------------- wait also the stop bit completion
	while (_TRMT == 0);
	
	// --------------------------------- send Speed/ CRC
	TransmitByte(Speed);
	TransmitByte(Crc);
	
	// --------------------------------- disable TX
	_UTXEN = 0;

	//Delay_us(100);
	COMM_DIR = 0;
	
	ToSendDataBuffer[0] = 0;  // command executed OK
}


