#ifndef COMMON_H
	#define COMMON_H
	
	
	// ==============================================================================================
	//  HARDWARE USAGE
	// ===============================================================================================
	    
    // INT1 =  -- unused -- 
    // INT2 =  PERIOD and USOUND_SENSOR
    // INT3 =  -- unused -- 
    
    // TMR1 = UART TimeOut 
    // TMR2 = PERIOD and USOUND_SENSOR
    // TMR3 = PWM and SERVO signals
    // TMR4 = FAST_COUNTER
    // TMR5 = -- unused -- 
    
    
	//  PIN      PORT	TRIS(1=input)	PULLUP(1=on)     CN         
	// ------------------------------------------------------------
	//	1		_RA0	_TRISA0			_CN2PUE			_CN2IE  
	//	2		_RA1	_TRISA1			_CN3PUE			_CN3IE  
	//	3		_RB0	_TRISB0			_CN4PUE			_CN4IE 
	//	4		_RB1	_TRISB1			_CN5PUE			_CN5IE  
	//	5		_RB2	_TRISB2			_CN6PUE			_CN6IE 
	//	6		_RB3	_TRISB3			_CN7PUE			_CN7IE 
	

	
	// ==============================================================================================
	//  INCLUDES
	// ==============================================================================================
	
	#include <p24FJ64GB002.h>
	#include <math.h>
	#include <stdio.h>
	//#include <stdlib.h>
	#include <string.h>
	#include "DEE Emulation 16-bit.h"
	
	
	// ==============================================================================================
	//  DEFINE
	// ==============================================================================================

	#define UART_RECEIVE_INTERRUPT 1
	#define UART1_RX_FIFO_SIZE 256

	#define CRYSTAL_4MHZ 0
	
	#define MASTER_LOGGER 0
	#define MASTER_SNIFFER 0
	
	#if (MASTER_SNIFFER)
		#define SNIFFED_SIZE 200
		#define SNIFFER_LOOPBACK 0
	#endif
	
	// ------------------------------------- check flags consistency
	#if (MASTER_LOGGER + MASTER_SNIFFER > 1)
		#error Flag inconsistence detected !!!
	#endif
	
	#define int8  signed char				//           -128	          127
	#define int16 signed int				//        -32,768	       32,767
	#define int32 signed long				// -2,147,483,648	2,147,483,647
	
	#define uint8  unsigned char			//				0			  255
	#define uint16 unsigned int				//        		0	       65,535
	#define uint32 unsigned long			//        		0	4,294,967,295
	
	// -------------------------------------------------- COMM DIRECTION ( 1 = TX / 0 = RX )
	//#define COMM_DIR LATBbits.LATB13
	#define COMM_DIR _RB13

	// -------------------------------------------------- STATUS LED
	#define LED_G  LATBbits.LATB5
	
	
	// -------------------------------------------------- CONSTANTS
	#define MAX_SLAVES 200
	#define MAX_PINS 12
	#define MASTER_PINS_COUNT 6
	
	#define INVALID_SLAVE_ID 255
	#define INVALID_SLAVE_TYPE 255
	
	// -------------------------------------------------- DELAYS
	#define FCY (32000000 / 2)				// instruction frequency (FCY = FOSC / 2)
	#define CYCLES_PER_MS (FCY / 1000) 		// instructions per millisecond 
	#define CYCLES_PER_US (FCY / 1000000) 	// instructions per microsecond 
	
	
	// ==============================================================================================
	//  ENUMS
	// ==============================================================================================
	
	// ------------------------------------------------------------------- Slave types
	enum SLAVE_TYPE
	{
		SLAVE_TYPE_CUSTOM			= 0,
		SLAVE_TYPE_CAPSENSOR		= 1,
		SLAVE_TYPE_SERVO			= 2,
		SLAVE_TYPE_GENERIC			= 3,
		SLAVE_TYPE_INOUT			= 4,
		SLAVE_TYPE_MASTER_PINS		= 5,
		SLAVE_TYPE_LOGGER			= 6,
		SLAVE_TYPE_SNIFFER			= 7,
		SLAVE_TYPE_UNKNOWN			= 255,
	};
	
	// ------------------------------------------------------------------- Pin types
	enum PIN_TYPE
	{
		UNUSED			= 0,
	
		DIG_OUT			= 1,
		PWM_8			= 2,
		PWM_16			= 3,
		SERVO_8			= 4,
		SERVO_16		= 5,
	
		DIG_IN			= 129,
		DIG_IN_PU		= 130,
	
		ADC_8 			= 131,
		ADC_16 			= 132,
		CAP_8 			= 133,
		CAP_16 			= 134,
		RES_8 			= 135,
		RES_16 			= 136,
	
		COUNTER 		= 140,
		COUNTER_PU 		= 141,
	
		FAST_COUNTER 	= 142,
		FAST_COUNTER_PU = 143,
		
		PERIOD 			= 144,
		PERIOD_PU 		= 145,
	
		USOUND_SENSOR	= 150,
		
		CAP_SENSOR 		= 160,
		
		I2C_SDA 		= 170,
		I2C_SCL 		= 171,
		
		MCA_8_BYTES		= 180,
		MCA_16_BYTES	= 181,
		MCA_32_BYTES	= 182,
	
		SNIFFER			= 190,
		
	//	VREF_POS 		= 180,
	//	VREF_NEG 		= 181,
	//
	//	UART_TX 		= 182,
	//	UART_RX 		= 183,
	//	UART_CTS 		= 184,
	//	UART_RTS 		= 185,
	//	I2C 			= 186,
	};
	
	// ------------------------------------------------------------------- Commands
	enum CMD
	{
		// ------------------------------------------ commands that master expands for the slaves
		CMD_RECOG_START 				= 254,
		CMD_RECOG 						= 253,
		CMD_FAST_DATA_EXCHANGE 			= 251,
	
	    // ------------------------------------------ commands that master repeats to slaves
		CMD_SETUP_SLAVE_PINS			= 249,
		CMD_SET_MASTER_NAME				= 248,
		CMD_GET_MASTER_NAME				= 247,
		CMD_SEND_VALUES 				= 246,
		CMD_GET_VALUES 					= 245,
		CMD_SEND_BYTES 					= 244,
		CMD_GET_BYTES 					= 243,
	
		// ------------------------------------------ commands to master only  ( < 200 )
		CMD_SPEED 						= 199,
		CMD_TEST_1 						= 99,
		CMD_NOCMD 						= 0
	};
	
	
	// ==============================================================================================
	//  VARS
	// ==============================================================================================
	

	
	
	
	// ==============================================================================================
	//   FUNCTION PROTOTYPES
	// ==============================================================================================
	
	void 	InitCrc();
	uint8 	AppendArrayToCrc(uint8* bytes, uint16 nBytes);
	uint8 	AppendByteToCrc(uint8 new_byte);
	
	void	TransmitByte(uint8 value);
	uint8	ReceiveByte();
	
	void	__delay32(unsigned long cycles);
	void 	Delay_ms(uint16 t);
	void	Delay_us(uint32 t);
	
	void	UserInit(void);
	void	InitUserAll();
	void	StopTimedOutputs();
	void 	SetInitialMasterPinTypes();
	void	EE_ReadAllParams();
	void	SetFactoryDefaults();
	void	EE_WriteAllParams();
	void	EE_InitialRead();
	
	void	UsbExchange();
	void	DelayLongTime_ms(uint16 t);
	uint8	GetMasterName();
	void	SetMasterName();
	
	void 	SetMasterInOutPins();
	void 	SendValuesToMasterPins(uint8* UsbBufIndex);
	void 	GetValuesFromMasterPins(uint8* UsbBufIndex);
	void 	MasterPins_SyncOutputPins();
	void	MasterPins_SyncInputPins();
	
	void	SendValuesToPhysicalSlave(uint8 SlaveId, uint8 nbytes, uint8 offset);
	void	GetValuesFromPhysicalSlave(uint8 SlaveId, uint8 nbytes, uint8 offset);
	void 	SendBytesToPhysicalSlave();
	void 	GetBytesFromPhysicalSlave();
	void 	TransmitComSpeedToPhysicalSlaves();
	
	void 	RecognizeSlaves();
	void	SetComSpeed(uint8 Speed);
	void	TransmitComSpeedToSlaves();
	void	ExecuteHostCommands();
	
	void	GenericTasks(void);
	
	void	LoggerManagement();
	void	SnifferManagement();
	
#endif

