

// ----------------------------------------------------- INCLUDES
#include "USB/usb.h"
#include "USB/usb_function_hid.h"
#include "usb_config.h"
#include "./common.h"




// ====================================================================
//  USB stack hardware selection options 
// ====================================================================
//  This section is the set of definitions required by the MCHPFSUSB
//  framework.  These definitions tell the firmware what mode it is
//  running in, and where it can find the results to some information
//  that the stack needs.
//  These definitions are required by every application developed with
//  this revision of the MCHPFSUSB framework.  Please review each
//  option carefully and determine which options are desired/required
//  for your application.

//#define USE_SELF_POWER_SENSE_IO
//#define tris_self_power     TRISAbits.TRISA2    // Input
//#define self_power          1

//#define USE_USB_BUS_SENSE_IO
//#define tris_usb_bus_sense  TRISBbits.TRISB5    // Input
#define USB_BUS_SENSE       1 

//Uncomment this to make the output HEX of this project 
//   to be able to be bootloaded using the HID bootloader
//#define PROGRAMMABLE_WITH_USB_HID_BOOTLOADER	

//If the application is going to be used with the HID bootloader
//  then this will provide a function for the application to 
//  enter the bootloader from the application (optional)
//#if defined(PROGRAMMABLE_WITH_USB_HID_BOOTLOADER)
//    #define EnterBootloader() __asm__("goto 0x400")
//#endif 




/** CONFIGURATION **************************************************/


// ------ CONFIG4 -------------------------------------------------------------------------
//
//   DSWDT Postscale Select:
//     DSWDTPS_DSWDTPS0     1:2 (2.1 ms)
//     DSWDTPS_DSWDTPS1     1:8 (8.3 ms)
//     DSWDTPS_DSWDTPS2     1:32 (33 ms)
//     DSWDTPS_DSWDTPS3     1:128 (132 ms)
//     DSWDTPS_DSWDTPS4     1:512 (528 ms)
//     DSWDTPS_DSWDTPS5     1:2,048 (2.1 seconds)
//     DSWDTPS_DSWDTPS6     1:8,192 (8.5 seconds)
//     DSWDTPS_DSWDTPS7     1:32,768 (34 seconds)
//     DSWDTPS_DSWDTPS8     1:131,072 (135 seconds)
//     DSWDTPS_DSWDTPS9     1:524,288 (9 minutes)
//     DSWDTPS_DSWDTPSA     1:2,097,152 (36 minutes)
//     DSWDTPS_DSWDTPSB     1:8,388,608 (2.4 hours)
//     DSWDTPS_DSWDTPSC     1:33,554,432 (9.6 hours)
//     DSWDTPS_DSWDTPSD     1:134,217,728 (38.5 hours)
//     DSWDTPS_DSWDTPSE     1:536,870,912 (6.4 days)
//     DSWDTPS_DSWDTPSF     1:2,147,483,648 (25.7 days)
//
//   Deep Sleep Watchdog Timer Oscillator Select:
//     DSWDTOSC_SOSC        DSWDT uses Secondary Oscillator (SOSC)
//     DSWDTOSC_LPRC        DSWDT uses Low Power RC Oscillator (LPRC)
//
//   RTCC Reference Oscillator  Select:
//     RTCOSC_LPRC          RTCC uses Low Power RC Oscillator (LPRC)
//     RTCOSC_SOSC          RTCC uses Secondary Oscillator (SOSC)
//
//   Deep Sleep BOR Enable bit:
//     DSBOREN_OFF          BOR disabled in Deep Sleep
//     DSBOREN_ON           BOR enabled in Deep Sleep
//
//   Deep Sleep Watchdog Timer:
//     DSWDTEN_OFF          DSWDT disabled
//     DSWDTEN_ON           DSWDT enabled


// ------ CONFIG3 -------------------------------------------------------------------------
//   Secondary Oscillator Pin Mode Select:
//     SOSCSEL_IO           SOSC pins have digital I/O functions (RA4, RB4)
//     SOSCSEL_LPSOSC       SOSC pins in Low-Power (low drive-strength) Oscillator Mode
//     SOSCSEL_SOSC         SOSC pins in Default (high drive-strength) Oscillator Mode

//   Voltage Regulator Wake-up Time Select:
//     WUTSEL_FST           Fast regulator start-up time used
//     WUTSEL_LEG           Default regulator start-up time used


// ------ CONFIG2 --------------------------------------------------------------------------
//
//   Primary Oscillator Select:
//     POSCMOD_EC           EC Oscillator mode selected
//     POSCMOD_XT           XT Oscillator mode selected
//     POSCMOD_HS           HS Oscillator mode selected
//     POSCMOD_NONE         Primary Oscillator disabled
//
//   I2C1 Pin Select bit:
//     I2C1SEL_SEC          Use alternate SCL1/SDA1 pins for I2C1
//     I2C1SEL_PRI          Use default SCL1/SDA1 pins for I2C1 
//
//   IOLOCK One-Way Set Enable:
//     IOL1WAY_OFF          The IOLOCK bit can be set and cleared using the unlock sequence
//     IOL1WAY_ON           Once set, the IOLOCK bit cannot be cleared
//
//   OSCO Pin Configuration:
//     OSCIOFNC_ON          OSCO pin functions as port I/O (RA3)
//     OSCIOFNC_OFF         OSCO pin functions as clock output (CLKO)
//
//   Clock Switching and Fail-Safe Clock Monitor:
//     FCKSM_CSECME         Sw Enabled, Mon Enabled
//     FCKSM_CSECMD         Sw Enabled, Mon Disabled
//     FCKSM_CSDCMD         Sw Disabled, Mon Disabled
//
//   Initial Oscillator Select:
//     FNOSC_FRC            Fast RC Oscillator (FRC)
//     FNOSC_FRCPLL         Fast RC Oscillator with Postscaler and PLL module (FRCPLL)
//     FNOSC_PRI            Primary Oscillator (XT, HS, EC)
//     FNOSC_PRIPLL         Primary Oscillator with PLL module (XTPLL, HSPLL, ECPLL)
//     FNOSC_SOSC           Secondary Oscillator (SOSC)
//     FNOSC_LPRC           Low-Power RC Oscillator (LPRC)
//     FNOSC_FRCDIV         Fast RC Oscillator with Postscaler (FRCDIV)
//
//   96MHz PLL Startup Select:
//     PLL96MHZ_OFF         96 MHz PLL Startup is enabled by user in software( controlled with the PLLEN bit)
//     PLL96MHZ_ON          96 MHz PLL Startup is enabled automatically on start-up
//
//   USB 96 MHz PLL Prescaler Select:
//     PLLDIV_NODIV         Oscillator input used directly (4 MHz input)
//     PLLDIV_DIV2          Oscillator input divided by 2 (8 MHz input)
//     PLLDIV_DIV3          Oscillator input divided by 3 (12 MHz input)
//     PLLDIV_DIV4          Oscillator input divided by 4 (16 MHz input)
//     PLLDIV_DIV5          Oscillator input divided by 5 (20 MHz input)
//     PLLDIV_DIV6          Oscillator input divided by 6 (24 MHz input)
//     PLLDIV_DIV8          Oscillator input divided by 8 (32 MHz input)
//     PLLDIV_DIV12         Oscillator input divided by 12 (48 MHz input)
//
//   Internal External Switchover:
//     IESO_OFF             IESO mode (Two-Speed Start-up) disabled
//     IESO_ON              IESO mode (Two-Speed Start-up) enabled
//




#if (MASTER_LOGGER)
	// ------------------------------------------------------------------------------------------------------------------------------
	//		START WITH SECONDARY 32 KHz OSCILLATOR
	// ------------------------------------------------------------------------------------------------------------------------------
	
	_CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)
	//
	_CONFIG3(WPFP_WPFP0 & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM & SOSCSEL_SOSC)
	//
	#if (CRYSTAL_4MHZ) 
		_CONFIG2(I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSECMD & PLL96MHZ_ON & PLLDIV_NODIV & IESO_OFF & FNOSC_SOSC)
	#else 
		_CONFIG2(I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSECMD & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF & FNOSC_SOSC)
	#endif
#else
	// ------------------------------------------------------------------------------------------------------------------------------
	//		START WITH PRIMARY 4 MHz OSCILLATOR
	// ------------------------------------------------------------------------------------------------------------------------------
	
	_CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)
	// 
	_CONFIG3(WPFP_WPFP0 & SOSCSEL_IO & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
	// 
	#if (CRYSTAL_4MHZ)
		_CONFIG2(POSCMOD_XT & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSECMD & FNOSC_PRIPLL & PLL96MHZ_ON & PLLDIV_NODIV & IESO_OFF)
	#else 
		_CONFIG2(POSCMOD_XT & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSECMD & FNOSC_PRIPLL & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF)
	#endif
#endif


// ==========================================================================
//  CODE PROTECT
//  Do not WRITE PROTECT because program memory is used as simulated eprom
// ==========================================================================
// ---------------------------------------------------------------------------------------------------- WatchDogTimer OFF - Protect OFF
_CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_OFF & ICS_PGx3 & GWRP_OFF & GCP_OFF & JTAGEN_OFF)
// ---------------------------------------------------------------------------------------------------- WatchDogTimer OFF - Protect ON
//_CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_OFF & ICS_PGx3 & GWRP_OFF & GCP_ON & JTAGEN_OFF)
// ---------------------------------------------------------------------------------------------------- WatchDogTimer ON - Protect ON
//_CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_ON & ICS_PGx3 & GWRP_OFF & GCP_ON & JTAGEN_OFF)



///** VARIABLES ******************************************************/
#pragma udata

unsigned char ReceivedDataBuffer[USB_EP0_BUFF_SIZE];
unsigned char ToSendDataBuffer[USB_EP0_BUFF_SIZE];

//
#pragma udata

USB_HANDLE USBOutHandle = 0;
USB_HANDLE USBInHandle = 0;

/** PRIVATE PROTOTYPES *********************************************/
//BOOL Switch2IsPressed(void);
//BOOL Switch3IsPressed(void);
static void InitializeSystem(void);
void UserInit(void);
//void YourHighPriorityISRCode();
//void YourLowPriorityISRCode();


/** DECLARATIONS ***************************************************/
#pragma code












// ******************************************************************************************************
// ************** USB Callback Functions ****************************************************************
// ******************************************************************************************************
// The USB firmware stack will call the callback functions USBCBxxx() in response to certain USB related
// events.  For example, if the host PC is powering down, it will stop sending out Start of Frame (SOF)
// packets to your device.  In response to this, all USB devices are supposed to decrease their power
// consumption from the USB Vbus to <2.5mA each.  The USB module detects this condition (which according
// to the USB specifications is 3+ms of no bus activity/SOF packets) and then calls the USBCBSuspend()
// function.  You should modify these callback functions to take appropriate actions for each of these
// conditions.  For example, in the USBCBSuspend(), you may wish to add code that will decrease power
// consumption from Vbus to <2.5mA (such as by clock switching, turning off LEDs, putting the
// microcontroller to sleep, etc.).  Then, in the USBCBWakeFromSuspend() function, you may then wish to
// add code that undoes the power saving things done in the USBCBSuspend() function.

// The USBCBSendResume() function is special, in that the USB stack will not automatically call this
// function.  This function is meant to be called from the application firmware instead.  See the
// additional comments near the function.

/******************************************************************************
 * Function:        void USBCBSuspend(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Call back that is invoked when a USB suspend is detected
 *
 * Note:            None
 *****************************************************************************/
void USBCBSuspend(void)
{
	//Example power saving code.  Insert appropriate code here for the desired
	//application behavior.  If the microcontroller will be put to sleep, a
	//process similar to that shown below may be used:
	
	//ConfigureIOPinsForLowPower();
	//SaveStateOfAllInterruptEnableBits();
	//DisableAllInterruptEnableBits();
	//EnableOnlyTheInterruptsWhichWillBeUsedToWakeTheMicro();	//should enable at least USBActivityIF as a wake source
	//Sleep();
	//RestoreStateOfAllPreviouslySavedInterruptEnableBits();	//Preferrably, this should be done in the USBCBWakeFromSuspend() function instead.
	//RestoreIOPinsToNormal();									//Preferrably, this should be done in the USBCBWakeFromSuspend() function instead.

	//IMPORTANT NOTE: Do not clear the USBActivityIF (ACTVIF) bit here.  This bit is 
	//cleared inside the usb_device.c file.  Clearing USBActivityIF here will cause 
	//things to not work as intended.	
	

    #if defined(__C30__)
        USBSleepOnSuspend();
    #endif
}


/******************************************************************************
 * Function:        void _USB1Interrupt(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called when the USB interrupt bit is set
 *					In this example the interrupt is only used when the device
 *					goes to sleep when it receives a USB suspend command
 *
 * Note:            None
 *****************************************************************************/
//void __attribute__ ((interrupt)) _USB1Interrupt(void)
//{
//    #if !defined(self_powered)
//        if(U1OTGIRbits.ACTVIF)
//        {
//            IEC5bits.USB1IE = 0;
//            U1OTGIEbits.ACTVIE = 0;
//            IFS5bits.USB1IF = 0;
//        
//            //USBClearInterruptFlag(USBActivityIFReg,USBActivityIFBitNum);
//            USBClearInterruptFlag(USBIdleIFReg,USBIdleIFBitNum);
//            //USBSuspendControl = 0;
//        }
//    #endif
//}


/******************************************************************************
 * Function:        void USBCBWakeFromSuspend(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The host may put USB peripheral devices in low power
 *					suspend mode (by "sending" 3+ms of idle).  Once in suspend
 *					mode, the host may wake the device back up by sending non-
 *					idle state signalling.
 *					
 *					This call back is invoked when a wakeup from USB suspend 
 *					is detected.
 *
 * Note:            None
 *****************************************************************************/
void USBCBWakeFromSuspend(void)
{
	// If clock switching or other power savings measures were taken when
	// executing the USBCBSuspend() function, now would be a good time to
	// switch back to normal full power run mode conditions.  The host allows
	// a few milliseconds of wakeup time, after which the device must be 
	// fully back to normal, and capable of receiving and processing USB
	// packets.  In order to do this, the USB module must receive proper
	// clocking (IE: 48MHz clock must be available to SIE for full speed USB
	// operation).
}


/********************************************************************
 * Function:        void USBCB_SOF_Handler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The USB host sends out a SOF packet to full-speed
 *                  devices every 1 ms. This interrupt may be useful
 *                  for isochronous pipes. End designers should
 *                  implement callback routine as necessary.
 *
 * Note:            None
 *******************************************************************/
void USBCB_SOF_Handler(void)
{
    // No need to clear UIRbits.SOFIF to 0 here.
    // Callback caller is already doing that.
}


/*******************************************************************
 * Function:        void USBCBErrorHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The purpose of this callback is mainly for
 *                  debugging during development. Check UEIR to see
 *                  which error causes the interrupt.
 *
 * Note:            None
 *******************************************************************/
void USBCBErrorHandler(void)
{
    // No need to clear UEIR to 0 here.
    // Callback caller is already doing that.

	// Typically, user firmware does not need to do anything special
	// if a USB error occurs.  For example, if the host sends an OUT
	// packet to your device, but the packet gets corrupted (ex:
	// because of a bad connection, or the user unplugs the
	// USB cable during the transmission) this will typically set
	// one or more USB error interrupt flags.  Nothing specific
	// needs to be done however, since the SIE will automatically
	// send a "NAK" packet to the host.  In response to this, the
	// host will normally retry to send the packet again, and no
	// data loss occurs.  The system will typically recover
	// automatically, without the need for application firmware
	// intervention.
	
	// Nevertheless, this callback function is provided, such as
	// for debugging purposes.
}


/*******************************************************************
 * Function:        void USBCBCheckOtherReq(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        When SETUP packets arrive from the host, some
 * 					firmware must process the request and respond
 *					appropriately to fulfill the request.  Some of
 *					the SETUP packets will be for standard
 *					USB "chapter 9" (as in, fulfilling chapter 9 of
 *					the official USB specifications) requests, while
 *					others may be specific to the USB device class
 *					that is being implemented.  For example, a HID
 *					class device needs to be able to respond to
 *					"GET REPORT" type of requests.  This
 *					is not a standard USB chapter 9 request, and 
 *					therefore not handled by usb_device.c.  Instead
 *					this request should be handled by class specific 
 *					firmware, such as that contained in usb_function_hid.c.
 *
 * Note:            None
 *******************************************************************/
void USBCBCheckOtherReq(void)
{
    USBCheckHIDRequest();
}//end


/*******************************************************************
 * Function:        void USBCBStdSetDscHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The USBCBStdSetDscHandler() callback function is
 *					called when a SETUP, bRequest: SET_DESCRIPTOR request
 *					arrives.  Typically SET_DESCRIPTOR requests are
 *					not used in most applications, and it is
 *					optional to support this type of request.
 *
 * Note:            None
 *******************************************************************/
void USBCBStdSetDscHandler(void)
{
    // Must claim session ownership if supporting this request
}//end


/*******************************************************************
 * Function:        void USBCBInitEP(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called when the device becomes
 *                  initialized, which occurs after the host sends a
 * 					SET_CONFIGURATION (wValue not = 0) request.  This 
 *					callback function should initialize the endpoints 
 *					for the device's usage according to the current 
 *					configuration.
 *
 * Note:            None
 *******************************************************************/
void USBCBInitEP(void)
{
    //enable the HID endpoint
    USBEnableEndpoint(HID_EP,USB_IN_ENABLED|USB_OUT_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
    //Re-arm the OUT endpoint for the next packet
    USBOutHandle = HIDRxPacket(HID_EP,(BYTE*)&ReceivedDataBuffer, USB_EP0_BUFF_SIZE);
}

/********************************************************************
 * Function:        void USBCBSendResume(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        The USB specifications allow some types of USB
 * 					peripheral devices to wake up a host PC (such
 *					as if it is in a low power suspend to RAM state).
 *					This can be a very useful feature in some
 *					USB applications, such as an Infrared remote
 *					control	receiver.  If a user presses the "power"
 *					button on a remote control, it is nice that the
 *					IR receiver can detect this signalling, and then
 *					send a USB "command" to the PC to wake up.
 *					
 *					The USBCBSendResume() "callback" function is used
 *					to send this special USB signalling which wakes 
 *					up the PC.  This function may be called by
 *					application firmware to wake up the PC.  This
 *					function should only be called when:
 *					
 *					1.  The USB driver used on the host PC supports
 *						the remote wakeup capability.
 *					2.  The USB configuration descriptor indicates
 *						the device is remote wakeup capable in the
 *						bmAttributes field.
 *					3.  The USB host PC is currently sleeping,
 *						and has previously sent your device a SET 
 *						FEATURE setup packet which "armed" the
 *						remote wakeup capability.   
 *
 *					This callback should send a RESUME signal that
 *                  has the period of 1-15ms.
 *
 * Note:            Interrupt vs. Polling
 *                  -Primary clock
 *                  -Secondary clock ***** MAKE NOTES ABOUT THIS *******
 *                   > Can switch to primary first by calling USBCBWakeFromSuspend()
 
 *                  The modifiable section in this routine should be changed
 *                  to meet the application needs. Current implementation
 *                  temporary blocks other functions from executing for a
 *                  period of 1-13 ms depending on the core frequency.
 *
 *                  According to USB 2.0 specification section 7.1.7.7,
 *                  "The remote wakeup device must hold the resume signaling
 *                  for at least 1 ms but for no more than 15 ms."
 *                  The idea here is to use a delay counter loop, using a
 *                  common value that would work over a wide range of core
 *                  frequencies.
 *                  That value selected is 1800. See table below:
 *                  ==========================================================
 *                  Core Freq(MHz)      MIP         RESUME Signal Period (ms)
 *                  ==========================================================
 *                      48              12          1.05
 *                       4              1           12.6
 *                  ==========================================================
 *                  * These timing could be incorrect when using code
 *                    optimization or extended instruction mode,
 *                    or when having other interrupts enabled.
 *                    Make sure to verify using the MPLAB SIM's Stopwatch
 *                    and verify the actual signal on an oscilloscope.
 *******************************************************************/
void USBCBSendResume(void)
{
    static WORD delay_count;
    
    USBResumeControl = 1;                // Start RESUME signaling
    
    delay_count = 1800U;                // Set RESUME line for 1-13 ms
    do
    {
        delay_count--;
    }while(delay_count);
    USBResumeControl = 0;
}


/*******************************************************************
 * Function:        BOOL USER_USB_CALLBACK_EVENT_HANDLER(
 *                        USB_EVENT event, void *pdata, WORD size)
 *
 * PreCondition:    None
 *
 * Input:           USB_EVENT event - the type of event
 *                  void *pdata - pointer to the event data
 *                  WORD size - size of the event data
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is called from the USB stack to
 *                  notify a user application that a USB event
 *                  occured.  This callback is in interrupt context
 *                  when the USB_INTERRUPT option is selected.
 *
 * Note:            None
 *******************************************************************/
BOOL USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, WORD size)
{
    switch(event)
    {
        case EVENT_CONFIGURED: 
            USBCBInitEP();
            break;
        case EVENT_SET_DESCRIPTOR:
            USBCBStdSetDscHandler();
            break;
        case EVENT_EP0_REQUEST:
            USBCBCheckOtherReq();
            break;
        case EVENT_SOF:
            USBCB_SOF_Handler();
            break;
        case EVENT_SUSPEND:
            USBCBSuspend();
            break;
        case EVENT_RESUME:
            USBCBWakeFromSuspend();
            break;
        case EVENT_BUS_ERROR:
            USBCBErrorHandler();
            break;
        case EVENT_TRANSFER:
            Nop();
            break;
        default:
            break;
    }      
    return TRUE; 
}








// =================================================================================================
//  Centralize initialization routine. 
//  All required USB initialization routines are called from here.
//
//  User application initialization routine should also be called from here.                  
// =================================================================================================

static void InitializeSystem(void)
{
	//On the PIC24FJ64GB002 Family of USB microcontrollers, the PLL will not power up and be enabled
	//by default, even if a PLL enabled oscillator configuration is selected (such as HS+PLL).
	//This allows the device to power up at a lower initial operating frequency, which can be
	//advantageous when powered from a source which is not gauranteed to be adequate for 32MHz
	//operation.  On these devices, user firmware needs to manually set the CLKDIV<PLLEN> bit to
	//power up the PLL.
    {
        unsigned int pll_startup_counter = 600;
        CLKDIVbits.PLLEN = 1;
        while(pll_startup_counter--);
    }

	//	The USB specifications require that USB peripheral devices must never source
	//	current onto the Vbus pin.  Additionally, USB peripherals should not source
	//	current on D+ or D- when the host/hub is not actively powering the Vbus line.
	//	When designing a self powered (as opposed to bus powered) USB peripheral
	//	device, the firmware should make sure not to turn on the USB module and D+
	//	or D- pull up resistor unless Vbus is actively powered.  Therefore, the
	//	firmware needs some means to detect when Vbus is being powered by the host.
	//	A 5V tolerant I/O pin can be connected to Vbus (through a resistor), and
	// 	can be used to detect when Vbus is high (host actively powering), or low
	//	(host is shut down or otherwise not supplying power).  The USB firmware
	// 	can then periodically poll this I/O pin to know when it is okay to turn on
	//	the USB module/D+/D- pull up resistor.  When designing a purely bus powered
	//	peripheral device, it is not possible to source current on D+ or D- when the
	//	host is not actively providing power on Vbus. Therefore, implementing this
	//	bus sense feature is optional.  This firmware can be made to use this bus
	//	sense feature by making sure "USE_USB_BUS_SENSE_IO" has been defined in the
	//	HardwareProfile.h file.    
//    #if defined(USE_USB_BUS_SENSE_IO)
//    tris_usb_bus_sense = INPUT_PIN; // See HardwareProfile.h
//    #endif
    
	//	If the host PC sends a GetStatus (device) request, the firmware must respond
	//	and let the host know if the USB peripheral device is currently bus powered
	//	or self powered.  See chapter 9 in the official USB specifications for details
	//	regarding this request.  If the peripheral device is capable of being both
	//	self and bus powered, it should not return a hard coded value for this request.
	//	Instead, firmware should check if it is currently self or bus powered, and
	//	respond accordingly.  If the hardware has been configured like demonstrated
	//	on the PICDEM FS USB Demo Board, an I/O pin can be polled to determine the
	//	currently selected power source.  On the PICDEM FS USB Demo Board, "RA2" 
	//	is used for	this purpose.  If using this feature, make sure "USE_SELF_POWER_SENSE_IO"
	//	has been defined in HardwareProfile.h, and that an appropriate I/O pin has been mapped
	//	to it in HardwareProfile.h.
//    #if defined(USE_SELF_POWER_SENSE_IO)
//    tris_self_power = INPUT_PIN;	// See HardwareProfile.h
//    #endif

    UserInit();
    
    USBDeviceInit();	//usb_device.c.  Initializes USB module SFRs and firmware
    					//variables to known states.
}







// =================================================================================================
// Function UserInit
// This routine should take care of all of the demo code initialization that is required.
// =================================================================================================

void UserInit(void)
{
	// ----------------------------------------------- All pins are digital
	AD1PCFG = 0b1111111111111111;
	
	// ----------------------------------------------- PORT A
	TRISA = 0b1111111111111111;								

	// ----------------------------------------------- PORT B (RB5 = LED) (RB13 = Direction)
	TRISB = 0b1101111111011111;

	LATA  = 0b0000000000000000;
	LATB  = 0b0000000000000000;

	// ----------------------------------------------- WeakPullUp on all unused pins
	_CN2PUE = 1;  // MasterPin 1
	_CN3PUE = 1;  // MasterPin 2
	_CN4PUE = 1;  // MasterPin 3
	_CN5PUE = 1;  // MasterPin 4
	_CN6PUE = 1;  // MasterPin 5
	_CN7PUE = 1;  // MasterPin 6
	
	_CN0PUE = 1;  // QZ2-32768
	_CN1PUE = 1;  // QZ2-32768
	
	_CN21PUE = 1; // AUX 2
	_CN22PUE = 1; // AUX 3
	_CN23PUE = 1; // AUX 4



	#if (MASTER_LOGGER)

		// -----------------------------------------------------------------------------------------
		//   TEST INITIAL BLINKINGS
		// -----------------------------------------------------------------------------------------
		
	//	_SOSCEN = 1;
	//// --------------------------------------------------------------------------------------- CLOCK SWITCHING
	//	__builtin_write_OSCCONH(0b000);    // NOSC[] = FRC
	//	__builtin_write_OSCCONH(0b101);    // NOSC[] = LPRC
	//	__builtin_write_OSCCONH(0b010);    // NOSC[] = XT, HS, EC
	//	__builtin_write_OSCCONH(0b100);    // NOSC[] = SOSC
	//	
	//	__builtin_write_OSCCONL(0x01);    // Start clock switching (OSWEN = 1)
	//	while (OSCCONbits.COSC != 0b000) {};
	//	while (OSCCONbits.LOCK != 1) {};
	//	while (_OSWEN) {};
	
		//ClockSwitch(0);	// NOSC[] = XT, HS, EC 
		//ClockSwitch(1);		// SOSC
		ClockSwitch(2);		// FRC
		//Delay_ms(50);
		//ClockSwitch(1);		// SOSC
		//ClockSwitch(3);		// LPRC
				
		while (1)
		{
			LED_G = 0;
			Delay_ms(5);
			LED_G = 1;
			Delay_ms(5);
		}
	#endif


		
	// --------------------------------------------------- used for FAST_COUNTER and FAST_COUNTER_PU
	T4CON = 0b1000000000000010;	// Timer4 - ON - 16bit - External clock from RP pin (on the rising edge)

	
	#if (MASTER_LOGGER)
		// ----------------------------------------------- PORT A
		// all input
		TRISA = 0b1111111111111111;
			
		// all input but: 
		// bit 13 ( Comm direction )
		// bit 5 (led green ) 
		TRISB = 0b1101111111011111; // TODO set to input all unused pins
	
		// ----------------------------------------------- NO WeakPullUp on used pins
		_CN7PUE = 0; // ADC_5 pin (Sensor value)
		_CN23PUE = 0;
		
		// ----------------------------------------------- RA0(buttons) as digital input without pullup
		_PCFG0 = 1;		
		_CN2PUE = 0;
	#endif

	#if (MASTER_SNIFFER && !SNIFFER_LOOPBACK)
		//	(bit 13) (Comm direction - INPUT)
		TRISB = 0b1111110011000000;
	#endif
	

	// =================================================================================================
	//   T1CON  
	// =================================================================================================

	//	bit 15 TON: Timer1 On bit
	//	  1 = Starts 16-bit Timer1
	//	  0 = Stops 16-bit Timer1

	//	bit 14 Unimplemented: Read as 0

	//	bit 13 TSIDL: Stop in Idle Mode bit
	//	  1 = Discontinue module operation when device enters Idle mode
	//	  0 = Continue module operation in Idle mode

	//	bit 12-7 Unimplemented: Read as 0

	//	bit 6 TGATE: Timer1 Gated Time Accumulation Enable bit
	//	  When TCS = 1: 
	//	    This bit is ignored.
	//	  When TCS = 0: 
	//	    1 = Gated time accumulation enabled
	//	    0 = Gated time accumulation disabled

	//	bit 5-4 TCKPS<1:0>: Timer1 Input Clock Prescale Select bits
	//	  11 = 1:256 
	//	  10 = 1:64
	//	  01 = 1:8 
	//	  00 = 1:1

	//	bit 3 Unimplemented: Read as 0

	//	bit 2 TSYNC: Timer1 External Clock Input Synchronization Select bit
	//	  When TCS = 1: 
	//	    1 =  Synchronize external clock input
	//	    0 =  Do not synchronize external clock input
	//	  When TCS = 0: 
	//	    This bit is ignored.

	//	bit 1 TCS: Timer1 Clock Source Select bit
	//	  1 =  External clock from T1CK pin (on the rising edge)
	//	  0 =  Internal clock (FOSC/2)

	//	bit 0 Unimplemented: Read as 0


	// ----------------------------------------------- TIMER 1 ( timeouts )
	//T1CON = 0b1000000000110000;	// Timer1 - ON - 16bit - prescaled_256 - from internal clock(16 MHz)
	//T1CON = 0b1000000000100000;	// Timer1 - ON - 16bit - prescaled_64 - from internal clock(16 MHz)
	//T1CON = 0b1000000000010000;	// Timer1 - ON - 16bit - prescaled_8 - from internal clock(16 MHz)
	//T1CON = 0b1000000000000000;	// Timer1 - 16bit - div 1 ( 16MHz ) - no-sync - from internal clock

	// ----------------- Used for Slave communication TIMEOUT
	// PPPPP
	//T1CON = 0b1000000000000000;	// Timer1 - 16bit - div 1 ( 16MHz ) - no-sync - from internal clock
	T1CON = 0b1000000000010000;	// Timer1 - ON - 16bit - prescaled_8 - from internal clock(16 MHz)
	
	
	// =================================================================================================
	//   TIMER 3 - ( SERVO and PWM )
	// =================================================================================================
	//T3CON = 0b1000000000000000;			// Timer3 - 16bit - Max speed - no-sync - from internal clock
	//T3CON = 0b1000000000010000;			// Timer3 - 16bit - div 8 ( 1MHz ) - no-sync - from internal clock
	T3CON = 0b1000000000000000;				// Timer3 - 16bit - div 1 ( 16MHz ) - no-sync - from internal clock

	//_T3IF = 0;	// Clear Timer3 Interrupt Flag
	//_T3IE = 1;	// Enable Timer3 interrupt
	
	StopTimedOutputs();
	

	// =================================================================================================
	//  TIMERS 23 45
	// =================================================================================================

	//T5CON = 0b1000000000110000;	// Timer5 - ON - 16bit - prescaled_256 - from internal clock(16 MHz)
	//T5CON = 0b1000000000100000;		// Timer5 - ON - 16bit - prescaled_64 - from internal clock(16 MHz)
	//T5CON = 0b1000000000010000;	// Timer5 - ON - 16bit - prescaled_8 - from internal clock(16 MHz)
	//T5CON = 0b1000000000000000;	// Timer5 - 16bit - div 1 ( 16MHz ) - no-sync - from internal clock



	// =================================================================================================
	//  UART configurations
	// =================================================================================================

	__builtin_write_OSCCONL(OSCCON & ~(1<<6));	// unlock the PPS registers
	RPINR18bits.U1RXR = 14;						// assign UART1 RX to pin RP14
	RPOR7bits.RP15R = 3;						// assign UART1 TX to pin RP15
	__builtin_write_OSCCONL(OSCCON | (1<<6));	// lock the PPS registers

//	U1STAbits.SPEN = 1 ; 	// enable rc6,rc7 as uart pins
//	U1STAbits.TX9 = 0 ; 	// Enable 8 bit transmission
//	U1STAbits.TXEN = 1 ; 	// Enable transmission
//	U1STAbits.SYNC = 0 ; 	// Asynchronous
//	
//	U1STAbits.RX9 = 0 ; 	// Enable 8 bit reception
//	U1STAbits.CREN = 1 ; 	// Enable reception
//	
//	PIE1bits.TXIE = 0 ; 	// Disable Transmission interrupt


	// ------------------------- enable Receive Interrupt

	#if UART_RECEIVE_INTERRUPT != 0
		//_U1RXIP = 1; 			//choose a priority
		_U1RXIF = 0;            //clear the flag
		_U1RXIE = 1;            //enable the interrupt
	#endif

	// --------------------------------------------------------- start with the default speed
	SetComSpeed(7);
	
	// ------------------------- enable UART and disable TX pin 
	_UARTEN = 1;
	_UTXEN = 0;
	
	// ------------------------- 00 = Interrupt is set when any character is received and transferred
	//                            from the RSR to the receive buffer.
	//                            Receive buffer has one or more characters.
	_URXISEL1 = 0;
	_URXISEL0 = 0;

	// ------------------------ 00 = Interrupt when a character is transferred to the Transmit Shift Register
	//                           (this implies there is at least one character open in the transmit buffer)
	// 							01 = Interrupt when the last character is shifted out of the Transmit Shift Register;
	//                           all transmit operations are completed
	_UTXISEL1 = 0;
	_UTXISEL0 = 1;
	
    
    //initialize the variable holding the handle for the last
    // transmission
    //USBOutHandle = 0;
    //USBInHandle = 0; 
    

	#if (MASTER_LOGGER)
		// =================================================================================================
		//  ADC Init
		// =================================================================================================
	   	_PCFG0 = 0;
		_PCFG11 = 0;
	
		_ASAM = 0; // sampling begins when SAMP bit is set
		_ADON = 1; // adc is on
	
		// --------------------------------------------------------------
		_SSRC = 0b111; 				// auto conv. starts after sampling
		AD1CSSL = 0; 				// no scanning
	
		AD1CON2 = 0; 				// use MUXA - References to AVss and AVdd	
		//_SMPI = 0b1111;  			// set interrupt flag after 16 convertions
	
		// --------------------------------------------------------------
		//  FCY = 16 MHz - TCY = 62.5 nS
		//  FCY = 32768 Hz - TCY = 30.518 uS
		//  "Sampling Time" must be greater than 750 nS  ( about 12 TAD )
		//  "A/D Clock Period" must be greater than 75 nS  ( about 2 TAD )
		//  conversion clock can be 1,2...64 * TCY
		// --------------------------------------------------------------
		//AD1CON3 = 0x1F0F;  // auto sample 31 TAD / ad conversion clock = 16 * TCY
		//AD1CON3 = 0x1F00;  // auto sample 31 TAD / ad conversion clock = 1 * TCY
		//AD1CON3 = 0x1F02;  // auto sample 31 TAD / ad conversion clock = 3 * TCY
		AD1CON3 = 0x1001;  // auto sample 16 TAD / ad conversion clock = 2 * TCY	// suggested by microchip - power floor = 110 watt
		//AD1CON3 = 0x100F;  // auto sample 16 TAD / ad conversion clock = 16 * TCY
		//AD1CON3 = 0x103F;  // auto sample 16 TAD / ad conversion clock = 64 * TCY	// medium time - power floor = 28 watt NOT WORKING
		//AD1CON3 = 0x1F3F;  // auto sample 31 TAD / ad conversion clock = 64 * TCY // slowest - power floor = 23 watt NOT WORKING
	
		//	AD1CON1 = 0x00E4; // integer format, auto conversion,sample after conversion ends             
		//
		//	AD1CON2 = 0x003C; // AVDD & AVSS as references,Disable Scan mode,interrupt after 16 sample/convert sequence,16 buffer levels, use MuxA              
		//
		//    AD1CON3 = 0x0D09;// AD Clock as derivative of system clock,13TAD,16 samples are collected in 1mSec,Fcy =4Mhz 1Tcy=0.26uS              
		//
		//	AD1CHS = 0x0005;// Positive sample input channel for MUX A to use AN5,Negative input channel for MUX A to use VR-
		//
		//	AD1PCFG = 0xFFDF; //AN5 analog pin
		//
		// 	AD1CSSL	= 0x0000; // Channel Scanning Disabled
	
		AD1CHS = 0x0005;	// Positive sample input channel for MUX A to use AN5,Negative input channel for MUX A to use VR-
		_PCFG5 = 0;			// AN5 analog pin
		_PCFG0 = 1;			// Digital In (TEST BUTTON)
	
		_SAMP = 1;
	#endif

	InitUserAll();
}










// ================================================================================
//    MAIN
// ================================================================================

int main(void)
{
    InitializeSystem();

    #if defined(USB_INTERRUPT)
        USBDeviceAttach();
    #endif

    while(1)
    {
	    #if defined(USB_POLLING)
        	USBDeviceTasks(); // Check bus status and service USB interrupts. (NOTE 1)
        #endif

		#if (MASTER_LOGGER)
			LoggerManagement();
		#endif

		#if (MASTER_SNIFFER)
			SnifferManagement();
		#endif
	
		UsbExchange(); 
	
		GenericTasks();
    }
}


// ========================================================================
//	  PROCESS USB EXCHANGES
// ========================================================================

void UsbExchange()
{
    if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1)) return;
    
    if(!HIDRxHandleBusy(USBOutHandle)) // Check if data was received from the host
    {
		ExecuteHostCommands();

        if(!HIDTxHandleBusy(USBInHandle))
        {
            USBInHandle = HIDTxPacket(HID_EP,(BYTE*)&ToSendDataBuffer,USB_EP0_BUFF_SIZE);
        }
   
        // ---------------------------- Re-arm the OUT endpoint for the next packet
        USBOutHandle = HIDRxPacket(HID_EP,(BYTE*)&ReceivedDataBuffer,USB_EP0_BUFF_SIZE);
    }
}    













// (NOTE 1)
// Interrupt or polling method.  
// If using polling, must call this function periodically.  
// This function will take care of processing and responding to SETUP transactions 
// (such as during the enumeration process when you first plug in).  
// USB hosts require that USB devices should accept and process SETUP packets in a timely fashion.  
// Therefore, when using polling, this function should be called frequently 
// (such as once about every 100 microseconds) 
// at any time that a SETUP packet might reasonably be expected to be sent by the host to your device.
//  In most cases, the USBDeviceTasks() function does not take very long to execute 
// (~50 instruction cycles) before it returns.





//	// ===============================================================================================
//	//   EERAM read all params
//	// ===============================================================================================
//
//	void EE_ReadAllParams()
//	{
//		uint16 i = 40;
//		// --------------------------------------
//		AverageSeconds = DataEERead(i); i+=2;
//		// --------------------------------------
//		float_uint16_t f; 
//		// --------------------------------------
//		f.uint16_var[0] = DataEERead(i); i+=2;
//		f.uint16_var[1] = DataEERead(i); i+=2;
//		TubeSens = f.float_var;
//		// --------------------------------------
//		f.uint16_var[0] = DataEERead(i); i+=2;
//		f.uint16_var[1] = DataEERead(i); i+=2;
//		TubeBkg = f.float_var;
//		// --------------------------------------
//		TubeDeadTime = DataEERead(i); i+=2;
//		// --------------------------------------
//		SoundType = DataEERead(i); i+=2;
//		// --------------------------------------
//		BackLightType = DataEERead(i); i+=2;
//	}	
//
//
//	// ===============================================================================================
//	//   EERAM write all params
//	// ===============================================================================================
//	
//	void EE_WriteAllParams()
//	{
//		uint16 i = 40;
//		// --------------------------------------
//		DataEEWrite(AverageSeconds, i); i+=2;
//		// --------------------------------------	
//		float_uint16_t f; 
//		// --------------------------------------
//		f.float_var = TubeSens;
//		DataEEWrite(f.uint16_var[0], i); i+=2;
//		DataEEWrite(f.uint16_var[1], i); i+=2;
//		// --------------------------------------
//		f.float_var = TubeBkg; 
//	 	DataEEWrite(f.uint16_var[0], i); i+=2;
//		DataEEWrite(f.uint16_var[1], i); i+=2;
//		// --------------------------------------
//		DataEEWrite(TubeDeadTime, i); i+=2;
//		// --------------------------------------	
//		DataEEWrite(SoundType, i); i+=2;
//		// --------------------------------------	
//		DataEEWrite(BackLightType, i); i+=2;
//	}

