
// ============================================================================
//  FILE: USER.C
// ============================================================================

// --------------------------------------------------- INCLUDES
#include "./USB/usb.h"
#include "./USB/usb_function_hid.h"

#include <math.h>
#include "./common.h"

// -------------------------------------------------- DEFINES
//
//#define TIMEOUT 62500 / 32  //  T1_freq / 32 mS    ( Timer1 freq = 16MHz / 256 = 62500 Hz )
//#define TIMEOUT 62500 / 15  //  T1_freq / 15 mS    ( Timer1 freq = 16MHz / 256 = 62500 Hz )

// -------------------------------------------------- VARS
extern unsigned char ReceivedDataBuffer[];
extern unsigned char ToSendDataBuffer[];

#if (MASTER_LOGGER)
	extern uint16 RMS_Value;
#endif

#if (MASTER_SNIFFER)
	extern uint8 SniffedBytes[SNIFFED_SIZE];
	extern uint8 SniffedFlags[SNIFFED_SIZE];
	extern uint8 SniffedIndex;
#endif

char   MasterName[17] = "";
uint16 RX_TimeOut;
uint16 TX_TimeOut;
uint8  RX_Error;

uint8  NumSlaves = 0;
uint8  NumPhysicalSlaves = 0;
uint32 ReadValue[MAX_SLAVES];
uint8  SlaveType[MAX_SLAVES];
uint8  PinType[MAX_SLAVES][MAX_PINS];

uint8  MTS_Bytes[MAX_SLAVES];
uint8  STM_Bytes[MAX_SLAVES];

uint8  gSpeed = 6;


// --------------------------------------------------
//#pragma code
//#pragma udata

// ==============================================================================================
//   DELAYS
// ==============================================================================================	
void Delay_ms(uint16 t)	// Max 65 sec
{	
	while(t--) __delay32(CYCLES_PER_MS);
}
void Delay_us(uint32 t)
{	
	__delay32(CYCLES_PER_US * (t - 3) + 5);
}



// =======================================================================================================
//    INTERRUPT TIMER
// =======================================================================================================
//uint16 OneSecondElapsedFlag = 0;
//uint16 OneSecondCounter;
//uint16 DelayCounter;
//
//void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _T5Interrupt(void)
//{
//	// --------------------------------------- ONE SECOND FLAG
//	if (++OneSecondCounter > 999)
//	{ 
//		OneSecondCounter = 0;
//		OneSecondElapsedFlag = 1;
//	}
//
//	DelayCounter++;
//	
//	// --------------------------------------- LCD square wave for negative voltage
//	LCD_SUPPLY = !LCD_SUPPLY;
//	
//	// --------------------------------------- reset the interrupt flag
//	_T5IF = 0;
//}



// =======================================================================================================
//    INTERRUPT ON ADC
// =======================================================================================================
#if (MASTER_LOGGER)
	void __attribute__((__interrupt__, __no_auto_psv__, __shadow__)) _ADC1Interrupt(void)
	{
		// --------------------------------------- reset the interrupt flag
		_AD1IF = 0;
	}
#endif
	
	
// ===============================================================================================
//   EERAM Initial read
// ===============================================================================================
void EE_InitialRead()
{
	GetMasterName();
}


// ===============================================================================================
//   INIT ALL
// ===============================================================================================
void InitUserAll()
{
	// ========================================================================
	//	ENSURE NO COMMANDS IN THE RECEIVE BUFFER
	// ========================================================================
	ReceivedDataBuffer[0] = CMD_NOCMD;
	
	// ========================================================================
	//	FLASH MEMORY INITIALIZATION
	// ========================================================================
	DataEEInit();
    dataEEFlags.val = 0;
    
    
    // ========================================================================
	//	VIRTUAL SLAVES
	// ========================================================================
	#if (MASTER_LOGGER)
		uint8 SlaveId = 0;
		NumSlaves = 1;
		
		SlaveType[SlaveId] = SLAVE_TYPE_LOGGER;

		PinType[SlaveId][0] = ADC_16;
		
		ToSendDataBuffer[SlaveId + 1] = SLAVE_TYPE_LOGGER;
	#elif (MASTER_SNIFFER)
		uint8 SlaveId = 0;
		NumSlaves = 1;
		
		SlaveType[SlaveId] = SLAVE_TYPE_SNIFFER;

		PinType[SlaveId][0] = SNIFFER;
		
		ToSendDataBuffer[SlaveId + 1] = SLAVE_TYPE_SNIFFER;
	#else
		uint8 SlaveId = 0;
		NumSlaves = 1;
		
		SlaveType[SlaveId] = SLAVE_TYPE_MASTER_PINS;

		SetInitialMasterPinTypes();
		
		ToSendDataBuffer[SlaveId + 1] = SLAVE_TYPE_MASTER_PINS;
	#endif
}



// ===============================================================================================
//   PINS
// ===============================================================================================
uint8 PinCommBytes(uint8 pinType)
{
	switch (pinType)
	{
		case UNUSED:
			return 0;
			
		case DIG_OUT:
		case PWM_8:
		case SERVO_8:
			return 1;
		
		case PWM_16:
		case SERVO_16:
			return 2;
			
		case DIG_IN:
		case DIG_IN_PU:
			return 1;
		
		case ADC_8:
		case CAP_8:
		case RES_8:
			return 1;
			
		case ADC_16:
		case CAP_16:
		case RES_16:
		case COUNTER:
		case COUNTER_PU:
		case FAST_COUNTER:
		case FAST_COUNTER_PU:
			return 2;
		
		case PERIOD:
		case PERIOD_PU:
			return 4;
		
		case USOUND_SENSOR:
			return 2;
		case CAP_SENSOR:
			return 3;
			
		case I2C_SDA:
			return 1;
		case I2C_SCL:
			return 1;
			
		case MCA_8_BYTES:
			return 8;
		case MCA_16_BYTES:
			return 16;
		case MCA_32_BYTES:	
			return 32;
			
//		case SNIFFER:
//			return 50;

		default:
			return 0;
	}	
}	

uint8 IsVirtualSlave(uint8 slaveType)
{
	switch (slaveType)
	{
		case SLAVE_TYPE_MASTER_PINS:
		case SLAVE_TYPE_LOGGER:
		case SLAVE_TYPE_SNIFFER:
			return TRUE;
		default:
			return FALSE;
	}
}	

uint8 SlavePinCount(uint8 slaveType)
{
	switch (slaveType)
	{
		case SLAVE_TYPE_CAPSENSOR:
			return 1;
		case SLAVE_TYPE_SERVO:
			return 10;
		case SLAVE_TYPE_GENERIC:
			return 12;
		case SLAVE_TYPE_INOUT:
			return 12;
		case SLAVE_TYPE_MASTER_PINS: // Virtual Slave
		case SLAVE_TYPE_LOGGER:
		case SLAVE_TYPE_SNIFFER:
		default:
			return 0;
	}
}		


// ===============================================================================================
//  HOST COMMAND - SETUP SLAVE PINS 
// ===============================================================================================
void SetupSlavePins()
{
	uint16 i;
	uint8 SlaveId = ReceivedDataBuffer[1];
	uint8 nbytes = ReceivedDataBuffer[2];
	
	uint8 CommandOK = TRUE;

	// -------------------------------------------------------- Physical Slaves only
	if (!IsVirtualSlave(SlaveType[SlaveId]))
	{
		TransmitByte(CMD_SETUP_SLAVE_PINS);
		TransmitByte(SlaveId);
		TransmitByte(nbytes);
		
		for (i = 0; i < nbytes; i++)
		{
			TransmitByte(ReceivedDataBuffer[3 + i]);
		}	
		InitCrc();
		uint8 crc = AppendArrayToCrc((uint8*)&ReceivedDataBuffer[0], 3 + nbytes);
		TransmitByte(crc);
		
		// ------------------------------------- receive the slaveID
		uint8 RX_Value = ReceiveByte();
		if (RX_Error) CommandOK = FALSE;
		if (RX_Value != SlaveId) CommandOK = FALSE;
		
		// ------------------------------------- receive the CRC
		RX_Value = ReceiveByte();
		if (RX_Error) CommandOK = FALSE;
		
		InitCrc();
		crc = AppendByteToCrc(SlaveId);
		if (RX_Value != crc) CommandOK = FALSE;
	}

	// -------------------------------------------------------- all the Slaves
	if (CommandOK)
	{
		MTS_Bytes[SlaveId] = 0;
		STM_Bytes[SlaveId] = 0;
		
		for (i = 0; i < nbytes; i++)
		{
			// ------------------------------------------------ set pintype
			uint8 pt = ReceivedDataBuffer[3 + i];
			PinType[SlaveId][i] = pt;
			
			// ------------------------------------------------ set MTS and STM bytes for this slave
			if (pt < 128) // 0..127 = "output" 
			{
				MTS_Bytes[SlaveId] += PinCommBytes(pt);
			}
			else
			{
				STM_Bytes[SlaveId] += PinCommBytes(pt);
			}	
		}
		
		// ---------------------------------------------------- setup MasterInOutPins
		if (SlaveId == 0)
		{
			SetMasterInOutPins();
		}	
	}	
	
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR		
}	


// ===============================================================================================
//  HOST COMMAND - GetMasterName
// ===============================================================================================
uint8 GetMasterName()
{
	uint16 i = 0;
	uint16 retval;
	
	ToSendDataBuffer[0] = 0;  // command executed OK
	
	while (1)
	{
		retval = DataEERead(i*2);
		if (retval == 0xFFFF)
		{
			MasterName[i] = '#';
			ToSendDataBuffer[0] = 1;  // command executed with ERROR
		}
		MasterName[i] = retval;
		Nop();
		Nop();
	
		ToSendDataBuffer[1 + i] = MasterName[i];
		
		if (i > 16)
		{
			// -----------------------------------------------------
			#if (MASTER_LOGGER) 	
				strcpy(MasterName, "Logger");
			#elif (MASTER_SNIFFER && SNIFFER_LOOPBACK)
				strcpy(MasterName, "Sniffer_LOOPBACK");
			#elif (MASTER_SNIFFER && !SNIFFER_LOOPBACK)
				strcpy(MasterName, "Sniffer");
			#else				
				strcpy(MasterName, "NoName");
			#endif
			// -----------------------------------------------------
			strcpy((char *)ToSendDataBuffer + 1, MasterName);
			// -----------------------------------------------------
			strcpy((char *)ReceivedDataBuffer + 1, MasterName);
			SetMasterName();
			// -----------------------------------------------------
			return 0;
		}		
		if (MasterName[i] == '\0') break;
		i++; 
	}
	return 1;
}	

// ===============================================================================================
//  HOST COMMAND - SetMasterName
// ===============================================================================================
void SetMasterName()
{
	uint16 i = 0;
	
	ToSendDataBuffer[0] = 0;  // command executed OK
	
	while (1)
	{
		MasterName[i] = ReceivedDataBuffer[1 + i];
		
		if (DataEEWrite(MasterName[i], i*2) != 0) ToSendDataBuffer[0] = 1;  // command executed with ERROR
		
		if (MasterName[i] == '\0') break;
		i++;
	}	
}


// ===============================================================================================
//  HOST COMMAND - FastDataExchange
// ===============================================================================================
void FastDataExchange(void)
{
	// -------------------------------------------------------------------------------------
	//  HOST TO MASTER TO SLAVES
	// -------------------------------------------------------------------------------------
	uint8 CommandOK = TRUE;
	uint8 UsbBufIndex;
	uint8 SlaveId;
	
	UsbBufIndex = 1;
	
	// ------------------------------------------------------------ Virtual Slaves
	SendValuesToMasterPins(&UsbBufIndex);

	// ------------------------------------------------------------ Physical slaves
	if (NumPhysicalSlaves > 0)
	{
		for (SlaveId = 1; SlaveId < NumSlaves; SlaveId++)
		{
			if (MTS_Bytes[SlaveId] > 0)
			{
				SendValuesToPhysicalSlave(SlaveId, MTS_Bytes[SlaveId], UsbBufIndex);
				UsbBufIndex += MTS_Bytes[SlaveId];
				if (ToSendDataBuffer[0] != 0) CommandOK = FALSE;
			}
		}
	}

	// -------------------------------------------------------------------------------------
	//   SLAVES TO MASTER TO HOST
	// -------------------------------------------------------------------------------------		
	UsbBufIndex = 1;

	// ------------------------------------------------------------ Virtual Slaves
	GetValuesFromMasterPins(&UsbBufIndex);

	// ------------------------------------------------------------ Physical slaves	
	if (NumPhysicalSlaves > 0)
	{
		for (SlaveId = 1; SlaveId < NumSlaves; SlaveId++)
		{
			if (STM_Bytes[SlaveId] > 0)
			{
				GetValuesFromPhysicalSlave(SlaveId, STM_Bytes[SlaveId], UsbBufIndex);
				UsbBufIndex += STM_Bytes[SlaveId];
				if (ToSendDataBuffer[0] != 0) CommandOK = FALSE;
			}		
		}
	}
	
	// ------------------------------------------------------------ set response
	if (CommandOK)
		ToSendDataBuffer[0] = 0; // command executed OK
	else
		ToSendDataBuffer[0] = 1; // command executed with ERROR	
}

// ===============================================================================================
//  HOST COMMAND - SendValuesToSlave   -  to PhysicalSlave or VirtualSlave(MasterPins)
// ===============================================================================================
void SendValuesToSlave()
{
	uint8 SlaveId = ReceivedDataBuffer[1];
	if (SlaveId == 0)
	{
		uint8 UsbBufIndex = 3;		
		SendValuesToMasterPins(&UsbBufIndex);
	}
	else
	{
		SendValuesToPhysicalSlave(SlaveId, ReceivedDataBuffer[2], 3);
	}	
}

// ===============================================================================================
//  HOST COMMAND - GetValuesFromSlave  -  from PhysicalSlave or VirtualSlave(MasterPins)
// ===============================================================================================
void GetValuesFromSlave()
{
	uint8 SlaveId = ReceivedDataBuffer[1];
	if (SlaveId == 0)
	{
		uint8 UsbBufIndex = 1;
		GetValuesFromMasterPins(&UsbBufIndex);
	}
	else
	{
		GetValuesFromPhysicalSlave(SlaveId, ReceivedDataBuffer[2], 1);
	}	
}


// ===============================================================================================
//   EXECUTE HOST COMMANDS
// ===============================================================================================
void ExecuteHostCommands(void)
{							
	switch (ReceivedDataBuffer[0])
	{
		// ------------------------------------------ SERVICE 
		case CMD_RECOG_START:
			RecognizeSlaves();
		break;
		case CMD_SETUP_SLAVE_PINS:
			SetupSlavePins();
		break;
		case CMD_GET_MASTER_NAME:
			GetMasterName();	
		break;
		case CMD_SET_MASTER_NAME:
			SetMasterName();	
		break;	
		// ------------------------------------------ COMMUNICATION
		case CMD_FAST_DATA_EXCHANGE:
			FastDataExchange();
		break;
		case CMD_SEND_VALUES:
			SendValuesToSlave();
		break;
		case CMD_GET_VALUES:
			GetValuesFromSlave();
		break;
		case CMD_SEND_BYTES:
			SendBytesToPhysicalSlave();
		break;
		case CMD_GET_BYTES:
			GetBytesFromPhysicalSlave();
		break;
		// ------------------------------------------ SPEED
		case CMD_SPEED: 
			TransmitComSpeedToPhysicalSlaves();	
		break;
	}
	ReceivedDataBuffer[0] = CMD_NOCMD;
}



// ===============================================================================================
//   GENERIC TASKS - called about every 10 uS
// ===============================================================================================
void GenericTasks(void)
{
	#if (MASTER_SNIFFER)
		
	#elif (MASTER_LOGGER)
    
		static uint16 BlinkLedsCounter = 0;
		if(BlinkLedsCounter-- == 0)
		{
			BlinkLedsCounter = 30000U;
	
			// ----------------------------------------------------------------------- clear watchdog timer
			ClrWdt();
	
			// ----------------------------------------------------------------------- show status on leds
		    if(USBSuspendControl == 1)
		    {
				LED_G = 0;
		    }
		    else
		    {
		   		if(USBDeviceState == CONFIGURED_STATE)	// blink GREEN
		        {
		            LED_G = !LED_G;
					//LED_G = !_RA0; 
		        }
				else
				{
					LED_G = 0;
		        }
			}
	 	}   
    
    #else
    
    	// -------------------------------------------------------------------------------------------
		//  BLINK LEDS
		// -------------------------------------------------------------------------------------------
	    static uint16 BlinkLedsCounter = 0;
		if(BlinkLedsCounter-- == 0)
		{
			BlinkLedsCounter = 10000U;
	
			// ----------------------------------------------------------------------- clear watchdog timer
			ClrWdt();
	
			// ----------------------------------------------------------------------- show status on led
		    if(USBSuspendControl == 1)
		    {
				LED_G = 0;
		    }
		    else
		    {
		   		if(USBDeviceState == CONFIGURED_STATE)	// blink GREEN
		        {
		            LED_G = !LED_G;    
		        }
		        else
		        {
					LED_G = 0;
		        }
			}
	    }
	#endif
}








